/*----------------------------------------------------------------------
  File    : Combined2.java
  Contents: split a (sub-)pointgon with triangles and paths
  Author  : Christian Borgelt
  History : 2006.01.18 file created (from Path.java)
            2006.01.19 split creation/traversal finished
            2006.01.20 construction of subproblems finished
            2006.01.21 hole distribution on subproblems finished
            2006.01.22 subproblem splitting debugged
            2006.01.27 triangle moved to abstract class Splitter
            2006.02.01 minor improvement in first init method
            2006.02.03 all triangle checks moved to valid(int,int,int)
            2006.02.05 made a subclass of Path rather than Splitter
            2006.02.06 function for getting the subproblem size improved
            2006.02.14 methods store(), retrieve() inherited from Path
----------------------------------------------------------------------*/
package pointgon;

/*--------------------------------------------------------------------*/
/** Class for splitting a (sub-)pointgon into (sub-)pointgons.
 *  @since  2006.01.27 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public class Combined2 extends Path {

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  private static final long serialVersionUID = 0x00010000;

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** direction of the subproblem key */
  private int keydir;
  /* In contrast to all other splitter classes, where the anchor of   */
  /* a subproblem with only one path is the left end of the path, in  */
  /* this class the anchor is always the clockwise end of the path.   */
  /* As a consequence the path can have two directions. For simpler   */
  /* checks the path direction in the subproblem is stored in keydir. */

  /*------------------------------------------------------------------*/
  /** Find vertex position in the current subproblem key.
   *  @param  a  first  vertex of triangle
   *  @param  b  second vertex of triangle
   *  @param  c  third  vertex of triangle
   *  @return whether the triangle is valid
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private int find (Vertex v)
  {                             /* --- find vertex position in key */
    int l, r, n;                /* indices for binary search */
    int c;                      /* result of comparison */

    l = 0; r = this.cnt;        /* do a binary search in the key */
    if (this.key[r-1] == null) r--;
    while (l < r) {             /* find the relative position of */
      n = (l+r) >> 1;           /* the vertex in the subproblem key */
      c = v.compareTo(this.key[n]);
      if (c == 0) return n;     /* if vertex has been found, abort */
      if (c == this.keydir) l = n+1;
      else                  r = n;
    }                           /* vertex v lies lexicographically */
    return l;                   /* between the vertices at l-1 and l */
  }  /* find() */

  /*------------------------------------------------------------------*/
  /** Check the added path edge.
   *  @return whether the added path edge is valid
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private boolean valid ()
  {                             /* --- check added path edge(s) */
    int    i;                   /* index of last vertex */
    Vertex v, p, q;             /* end vertices of edge(s) */

    if (this.plen > 2) {        /* if there is more than one hole */
      p = this.path[this.plen-3];  /* get and check the edge that */
      q = this.path[this.plen-2];  /* was added to the sequence */
      return (this.wgts[p.id][q.id] >= 0);
    }                           /* return whether edge is valid */
    v = this.path[0];           /* get the tip of the triangle */
    if (this.pdir != this.keydir) {
      p = this.key[0];          /* if path starts at start of key */
      q = this.key[1]; }        /* get the first two key vertices */
    else {                      /* if path starts at end of key */
      q = this.key[i = this.cnt-1];
      if (q == null) q = this.key[--i];
      p = this.key[i-1];        /* get the last  two key vertices */
    }                           /* (base of the triangle to be cut) */
    return this.valid(p, q, v); /* return whether triangle is valid */
  }  /* valid() */

  /*------------------------------------------------------------------*/
  /** Check whether a hole is inside a subproblem.
   *  @param  hole  the hole to check
   *  @param  key   the key  of the subproblem
   *  @param  cnt   the size of the subproblem
   *  @return whether the hole is inside the subproblem
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private boolean inside (Vertex hole, Vertex[] key, int cnt)
  {                             /* --- check whether a hole is inside */
    int    i, r, n, c = 0;      /* loop variables, buffer, counter */
    Vertex p, q, v;             /* perimeter vertices of subproblem */

    if (key[cnt-1] == null)     /* get the number of key vertices */
      cnt--;                    /* without the perimeter section */
    n = this.verts.length;      /* get modulo for id computations */
    p = v = key[0]; i = 1;      /* traverse vertices on the perimeter */
    do {                        /* of the given subproblem */
      q = (i < cnt) ? key[i++] : this.verts[(p.id +n +this.dir) % n];
      r = hole.horzLineCuts(p, q); /* check whether horz. line cuts */
      if (r == 0) return false; /* if vertex is on the edge, abort */
      if (r >  0) c++;          /* if horz. line cuts to the right */
      p = q;                    /* of the vertex, count the edge */
    } while (q != v);           /* while not back at the anchor */
    /* The idea of the above check is to shoot a straight line from   */
    /* the given point to the right and to count how many times this  */
    /* line cuts the perimeter. If the number is even, the point lies */
    /* outside the perimeter, if it is odd, the point lies inside.    */
    return (c & 1) != 0;        /* check the number of intersections */
  }  /* inside() */             /* (odd: inside, even: outside) */

  /*------------------------------------------------------------------*/
  /** Get the full size of a subproblem.
   *  @param  key   the key of the subproblem
   *  @param  cnt   the size of the subproblem
   *  @return the full size of a subproblem
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected int size (Vertex[] key, int cnt)
  {                             /* --- get full size of subproblem */
    int    k, n;                /* full size, number of vertices */
    Vertex p, q;                /* two corners of the triangle */

    if (cnt < 3) return 0;      /* check for an empty subproblem */
    if (key[--cnt] != null)     /* if there is no perimeter section, */
      return cnt+1;             /* return the key size */
    k = key[0].id -key[cnt-1].id;
    if (this.dir < 0) k = -k;   /* compute edges in perim. section */
    if (k <= 0) k += this.verts.length;
    k += cnt-1;                 /* compute the total number of edges */
    if ((k > 3) || (this.trtab == null))
      return k;                 /* check size and precomputed info. */
    p = key[1]; q = key[2];     /* get the two triangle corners */
    if (q == null) {            /* if there is a perimeter section */
      n = this.verts.length;    /* get modulo for id computation */
      q = this.verts[(p.id +n +this.dir) % n];
    }                           /* get vertex in perimeter section */
    return (this.trtab[key[0].id][p.id][q.id] != 0)
         ? 0 : k;               /* check for an empty triangle and */
  }  /* size() */               /* return the corr. subproblem size */

  /*------------------------------------------------------------------*/
  /** Create a subproblem splitter.
   *  @param  mwt   the minimum weight traingulation object
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Combined2 (MWT mwt)
  { super(mwt); this.triangle = new Vertex[3]; }

  /*------------------------------------------------------------------*/
  /** Initialize the subproblem splitter.
   *  @param  key      the key  of the subproblem
   *  @param  cnt      the size of the subproblem
   *  @param  holes    the holes of the subproblem
   *  @param  holecnt  the number of holes of the subproblem
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void init (Vertex[] key,   int cnt,
                       Vertex[] holes, int holecnt)
  {                             /* --- init. a subproblem splitter */
    Vertex v;                   /* end vertex of the key */

    this.key     = key;         /* note the key of the subproblem */
    this.cnt     = cnt;         /* and its lexicographic direction */
    this.keydir  = key[1].compareTo(key[0]);
    this.holes   = holes;       /* note the holes that are (possibly) */
    this.holecnt = holecnt;     /* inside this subproblem */
    this.pdir    = 0;           /* init. the split state */
    this.state   = --cnt;       /* (no splits created yet) */
    this.bestlen = 0;           /* init. the best path length */
    v = key[cnt];               /* get and check the last key vertex */
    if (v == null) v = key[--this.state];
    /* Note that the above statements must be executed before the    */
    /* number of holes is checked, because they may adapt the state. */
    if      (holecnt     <= 0) this.lftpos = this.rgtpos = 0;
    else if (this.keydir >  0) this.getpos(key[0], v);
    else                       this.getpos(v, key[0]);
  }  /* init() */               /* get hole positions for paths */

  /*------------------------------------------------------------------*/
  /** Initialize a subproblem splitter.
   *  @param  key      the key  of the subproblem
   *  @param  cnt      the size of the subproblem
   *  @param  spobj    the splitter object
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void init (Vertex[] key, int cnt, Object spobj)
  {                             /* --- init. a subproblem splitter */
    int    i;                   /* key index */
    Vertex v;                   /* tip of a triangle */

    this.key    = key;          /* note the key of the subproblem */
    this.cnt    = cnt;          /* and its lexicographic direction */
    this.keydir = key[1].compareTo(key[0]);
    if (spobj instanceof Vertex) {
      this.plen    = 1;         /* if split object is a triangle, */
      this.pdir    = 0;         /* get the tip of the triangle */
      this.path[0] = v = (Vertex)spobj;
      i = this.find(v);         /* store tip and find its position */
      this.state = ((i >= 0) && (i < cnt) && (v == key[i]))
                 ? i : -i; }    /* find position in subproblem key */
    else {                      /* if the split object is a path */
      this.plen = ((Vertex[])spobj).length;
      System.arraycopy(spobj, 0, this.path, 0, this.plen);
      this.pdir = this.path[0].compareTo(key[0]);
    }                           /* copy the splitting path */
  }  /* init() */

  /*------------------------------------------------------------------*/
  /** Create next subproblem split.
   *  @return whether a new split was created
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected boolean next ()
  {                             /* --- create next subproblem split */
    int     i, k, e, n;         /* indices, number of vertices */
    Vertex  v, p, q;            /* key/path/perimeter vertices */
    boolean valid;              /* whether prec. sequence was valid */

    e = this.cnt-1;             /* get index of last key vertex */
    if (this.key[e] == null) e--;

    /* --- cut off holes of the key --- */
    if ((this.pdir == 0)        /* if no paths and no triangles */
    &&  (this.state > 0)) {     /* have been processed up to now */
      p = this.key[this.state-1];
      v = this.key[this.state]; /* get preceding path edge */
      while (--this.state > 0){ /* cut vertex search loop */
        q = v; v = p;           /* shift cut along the path */
        p = this.key[this.state-1];
        if ((this.wgts[p.id][q.id] >= 0)
        &&  this.valid(p, v, q)) {
          this.path[0] = v;     /* if the cut-off triangle is valid, */
          this.plen    = 1;     /* store the hole that is the tip */
          return true;          /* and return 'next split created' */
        }                       /* (edge cutting off a key hole) */
      }
      this.path[0] = this.key[e];
      this.idxs[0] = -1;        /* switch to triangles over edges */
    }                           /* (start with perimeter vertices) */

    /* --- build triangles over edges --- */
    if (this.pdir == 0) {       /* if no paths processed yet, */
      n = this.verts.length;    /* get modulo for id computations */
      v = this.path[0];         /* get previous triangle tip */
      while (true) {            /* triangle tip search loop */
        if (this.idxs[0] < 0) { /* if in the perimeter vertices */
          v = this.verts[(v.id +n +this.dir) % n];
          if (v == this.key[0]){/* if wrapped round to the anchor */
            this.idxs[0] = this.rgtpos;
            continue;           /* all perimeter vertices are done, */
          }                     /* so switch to traversing the holes */
          i = k = this.find(v); /* find position of tip in the key */
          if      (k <= 0) k++;    /* adapt extreme */
          else if (k >  e) k--; }  /* key index values */
        else {                  /* if in the hole vertices */
          i = --this.idxs[0];   /* get the index of the next hole */
          if (i < this.lftpos)  /* if all holes in key range have */
            break;              /* been processed, abort the loop */
          v = this.holes[i];    /* get the next hole vertex */
          i = k = this.find(v); /* find position of tip in the key */
        }                       /* (defines the triangle base) */
        p = this.key[k-1];      /* get the base vertices */
        q = this.key[k];        /* of the triangle to cut */
        if (this.valid(p, q, v)) {
          this.path[0] =  v;    /* if the triangle is valid, */
          this.plen    =  1;    /* store its tip and the index */
          this.state   = -i;    /* note start vertex of the edge */
          return true;          /* and return 'next split created' */
        }                       /* (triangle built over an edge) */
      }
      if (this.holecnt <= 0)    /* if there are no hole vertices, */
        return false;           /* all splits have been generated */
      do {                      /* hole sequence generation loop */
        if (!this.nextHoles(false))
          return false;         /* check for another hole sequence */
      } while (!this.valid());  /* while hole sequence is invalid */
      this.path[this.plen-1] = p = this.key[e];
    }                           /* (switch to splits with paths) */

    /* --- attach a path to a triangle --- */
    n = this.verts.length;      /* get modulo for id computations */
    q = this.path[this.plen-1]; /* and the previous path end */
    while (true) {              /* path generation loop */
      q = this.verts[(q.id +n +this.dir) % n];
      if (q != this.key[0]) {   /* if there is another end vertex */
        p = this.path[this.plen-2];
        if ((this.wgts[p.id][q.id] < 0)
        ||  (q.compareTo(p) != this.pdir))
          continue;             /* if the edge is invalid, skip it */
        this.path[this.plen-1] = q;   /* store the next end vertex */
        return true;            /* and return 'next split created' */
      }                         /* (triangle with attached path) */
      valid = true;             /* preceding hole sequence was valid */
      do {                      /* hole sequence generation loop */
        if (!this.nextHoles(valid))
          return false;         /* check for another hole sequence */
        valid = this.valid();   /* and then check its validity */
      } while (!valid);         /* while hole sequence is invalid */
      this.path[this.plen-1] = q = this.key[e];
    }                           /* reinitialize the path end */
  }  /* next() */

  /*------------------------------------------------------------------*/
  /** Split a subproblem with the current split.
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void split ()
  {                             /* --- split subproblem with path */
    int    i, k, e, n;          /* indices, numbers of vertices */
    Vertex v, p, q;             /* key/path/perimeter vertices */

    /* --- edge cutting off a key hole --- */
    if ((this.pdir == 0)        /* if the split is no path and */
    &&  (this.state > 0)) {     /* no triangle built over an edge */
      i = this.state;           /* get index of cut-off vertex */
      if (this.keydir > 0) {    /* if the triangle points downward */
        this.triangle[0] = this.key[i-1];
        this.triangle[1] = this.key[i];
        this.triangle[2] = this.key[i+1]; }
      else {                    /* if the triangle points upward */
        this.triangle[0] = this.key[i+1];
        this.triangle[1] = this.key[i-1];
        this.triangle[2] = this.key[i];
      }                         /* (make leftmost vertex anchor) */
      k = this.cnt-1;           /* get number of rem. key vertices */
      System.arraycopy(this.key, 0,   this.lftkey, 0, i);
      System.arraycopy(this.key, i+1, this.lftkey, i, k-i);
      this.lftcnt = k;          /* copy the remaining key vertices */
      this.rgtcnt = 0; return;  /* left key has one vertex less, */
    }                           /* right subproblem key is empty */

    n = this.verts.length;      /* get modulo for id computations */
    p = this.key[0];            /* and first and last key vertex */
    q = this.key[e = this.cnt-1];
    if (q == null) q = this.key[--e];

    /* --- pure triangle built over an edge --- */
    if (this.pdir == 0) {       /* if the split is no path */
      v = this.path[0];         /* get the triangle tip and */
      i = -this.state;          /* the number of preceding vertices */
      if      (i <= 0) {        /* if before the first key vertex */
        if (this.keydir > 0) {  /* if the triangle points upwards */
          this.triangle[0] = v; /* make triangle tip the anchor, */
          this.triangle[1] = p; /* then do counterclockwise walk */
          this.triangle[2] = this.key[1]; }
        else {                  /* if the triangle points downwards */
          this.triangle[0] = this.key[1];
          this.triangle[1] = v; /* make 2nd key vertex the anchor, */
          this.triangle[2] = p; /* then do counterclockwise walk */
        } }                     /* (ensure unique triangle coding) */
      else if (i > e) {         /* if after the last key vertex */
        if (this.keydir > 0) {  /* if the triangle points upwards */
          this.triangle[0] = this.key[e-1];
          this.triangle[1] = q; /* make leftmost vertex the anchor, */
          this.triangle[2] = v;}/* then do counterclockwise walk */
        else {                  /* if the triangle points downwards */
          this.triangle[0] = v; /* make triangle tip the anchor */
          this.triangle[1] = this.key[e-1];
          this.triangle[2] = q; /* then do counterclockwise walk */
        } }                     /* (ensure unique triangle coding) */
      else {                    /* if truly above a key edge */
        if (this.keydir > 0) {  /* if the triangle points upwards */
          this.triangle[0] = this.key[i-1];
          this.triangle[1] = this.key[i];
          this.triangle[2] = v;}/* tip is ccw-wise last vertex */
        else {                  /* if the triangle points downwards */
          this.triangle[0] = this.key[i];
          this.triangle[1] = v; /* tip is ccw-wise second vertex */
          this.triangle[2] = this.key[i-1];
        }                       /* (ensure unique triangle coding) */
        if (v.id >= n) {        /* if the tip is a hole vertex */
          this.lftkey[i] = v;   /* store the triangle tip */
          System.arraycopy(this.key, 0, this.lftkey, 0,   i);
          System.arraycopy(this.key, i, this.lftkey, i+1, this.cnt-i);
          this.lftcnt = this.cnt+1;
          this.rgtcnt = 0;      /* surround triangle tip with old key */
          return;               /* (insert tip at lexicogr. position) */
        }                       /* left key has one vertex more, */
      }                         /* right subproblem key is empty */
      if (((i <= 1) || (e < 2)) /* if cutting off the anchor */
      &&  (v.id == (p.id +n -this.dir) % n)) {
        System.arraycopy(this.key, 1, this.rgtkey, 1, e);
        this.rgtkey[0] = v;     /* replace first key vertex */
        this.rgtcnt    = e+1;   /* with the triangle tip */
        if (v.id != (q.id +n +this.dir) % n)
          this.rgtkey[this.rgtcnt++] = null;
        this.lftcnt = 0;        /* add perimeter section if necessary */
        return;                 /* right subproblem has same size, */
      }                         /* left subproblem key is empty */
      if (((i >= e) || (e < 2)) /* if cutting off the key end */
      &&  (v.id == (q.id +n +this.dir) % n)) {
        System.arraycopy(this.key, 0, this.lftkey, 0, e);
        this.lftkey[e] = v;     /* replace last key vertex */
        this.lftcnt    = e+1;   /* with the triangle tip */
        if (v.id != (p.id +n -this.dir) % n)
          this.lftkey[this.lftcnt++] = null;
        this.rgtcnt = 0;        /* add perimeter section if necessary */
        return;                 /* left subproblem has same size, */
      }                         /* right subproblem key is empty */
      if      (i <= 0) i++;     /* adapt extreme */
      else if (i >  e) i--;     /* key index values */
      System.arraycopy(this.key, 0, this.lftkey, 0, i);
      this.lftkey[i] = v;       /* store the triangle tip after */
      this.lftcnt    = i+1;     /* the beginning of the old key */
      if (v.id != (p.id +n -this.dir) % n)
        this.lftkey[this.lftcnt++] = null;
      k = e+1-i;                /* get number of rem. key vertices */
      System.arraycopy(this.key, i, this.rgtkey, 1, k);
      this.rgtkey[0] = v;       /* store the triangle tip before */
      this.rgtcnt    = k+1;     /* the remaining part of the key */
      if (v.id != (q.id +n +this.dir) % n)
        this.rgtkey[this.rgtcnt++] = null;
      return;                   /* the key sizes are the part sizes */
    }                           /* plus tip plus possible per. sect. */

    /* --- triangle with attached path --- */
    k = this.plen;              /* get the length of the path */
    if (this.pdir != this.keydir) {
      this.lftkey[0] = p;       /* if path starts before the key */
      System.arraycopy(this.path, 0, this.lftkey, 1, k);
      this.lftcnt = k+1;        /* left subproblem key is the path */
      if (this.lftkey[0].id != (this.lftkey[k].id +n +this.dir) % n)
        this.lftkey[this.lftcnt++] = null;
      for (k = 0, i = this.plen; --i >= 0; )
        this.rgtkey[k++] = this.path[i];
      System.arraycopy(this.key, 1, this.rgtkey, k, e);
      this.rgtcnt = k +e;       /* right key is reversed path and key */
      if (this.rgtkey[0].id != (q.id +n +this.dir) % n)
        this.rgtkey[this.rgtcnt++] = null;
      if (this.pdir > 0) {      /* if the path is left to right */
        this.triangle[0] = this.key[1];
        this.triangle[1] = this.path[0];
        this.triangle[2] = p; } /* second key vertex is leftmost */
      else {                    /* if the path is right to left */
        this.triangle[0] = this.path[0];
        this.triangle[1] = p;   /* first path vertex is leftmost */
        this.triangle[2] = this.key[1];
      } }                       /* (ensure unique triangle coding) */
    else {                      /* if path starts after end of key */
      System.arraycopy(this.key,  0, this.lftkey, 0, e);
      System.arraycopy(this.path, 0, this.lftkey, e, k);
      this.lftcnt = e +k;       /* left key is old key and path */
      if (this.lftkey[0].id != (this.path[k-1].id +n +this.dir) % n)
        this.lftkey[this.lftcnt++] = null;
      for (k = 0, i = this.plen; --i >= 0; )
        this.rgtkey[k++] = this.path[i];
      this.rgtkey[k++] = q;     /* right subproblem is reversed path */
      this.rgtcnt = k;          /* add perimeter section if necessary */
      if (this.rgtkey[0].id != (q.id +n +this.dir) % n)
        this.rgtkey[this.rgtcnt++] = null;
      if (this.pdir > 0) {      /* if the path is left to right */
        this.triangle[0] = this.key[e-1];
        this.triangle[1] = q;   /* semi-last key vertex is leftmost */
        this.triangle[2] = this.path[0]; }
      else {                    /* if the path is right to left */
        this.triangle[0] = this.path[0];
        this.triangle[1] = q;   /* first path vertex is leftmost */
        this.triangle[2] = this.key[e-1];
      }                         /* (ensure unique triangle coding) */
    }
  }  /* split() */

  /*------------------------------------------------------------------*/
  /** Get the current triangle.
   *  @return always <code>null</code>
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Vertex[] triangle ()
  { return this.triangle; }

  /*------------------------------------------------------------------*/
  /** Compute the weight of the current split.
   *  @return the weight of the current split
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected double weight ()
  {                             /* --- compute weight of curr. split */
    int    i, e, n;             /* indices, number of vertices */
    Vertex v, p, q;             /* triangle tip, key vertices */
    double wgt;                 /* weight/length of the path */

    v = this.path[0];           /* get the triangle tip/path start */
    if (this.plen <= 1) {       /* if split with only a triangle */
      i = this.state;           /* get type/position of triangle tip */
      if (i > 0)                /* if cutting off a hole of the key */
        return this.wgts[this.key[i-1].id][v.id]
             + this.wgts[this.key[i+1].id][v.id];
      n = this.verts.length;    /* get modulo for id computations */
      i = -i;                   /* and the tip position in the key */
      p = this.key[0];          /* get first and last key vertex */
      q = this.key[e = this.cnt-1];
      if (q == null) q = this.key[--e];
      if (((i <= 1) || (e < 2)) /* if triangle cuts off the anchor */
      &&  (v.id == (p.id +n -this.dir) % n))
        return this.wgts[p.id][v.id]
             + this.wgts[p.id][this.key[  1].id];
      if (((i >= e) || (e < 2)) /* if triangle cuts off the key end */
      &&  (v.id == (q.id +n +this.dir) % n))
        return this.wgts[q.id][v.id]
             + this.wgts[q.id][this.key[e-1].id];
      if      (i <= 0) i++;     /* adapt extreme */
      else if (i >  e) i--;     /* key index values */
      return this.wgts[this.key[i].id][this.key[i-1].id];
    }                           /* compute the split weight */
    if (this.pdir !=this.keydir)/* if triangle is at start of key */
      wgt = this.wgts[this.key[0].id][this.key[1].id];
    else {                      /* if triangle is at end   of key */
      p = this.key[e = this.cnt-1];
      if (p == null) p = this.key[--e];
      wgt = this.wgts[p.id][this.key[e-1].id];
    }                           /* get weight of triangle base */
    for (i = this.plen; --i > 0; )
      wgt -= this.wgts[this.path[i].id][this.path[i-1].id];
    return wgt;                 /* subtract the edge lengths and */
  }  /* weight() */             /* return the computed split weight */

  /*------------------------------------------------------------------*/
  /** Get the edges of the current split.
   *  @return the edges of the current split as an array
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected Edge[] edges ()
  {                             /* --- get edges of current split */
    int    i, e, n;             /* indices, number of vertices */
    Vertex v, p, q;             /* triangle tip, key vertices */
    Edge[] edges;               /* edges of current path */

    /* --- pure triangle split --- */
    if (this.pdir == 0) {       /* if split with only a triangle */
      i = this.state;           /* get type/position of triangle tip */
      if (i > 0) {              /* if cutting off a hole of the key, */
        edges    = new Edge[1]; /* the triangle base is the only edge */
        edges[0] = new Edge(this.key[i-1], this.key[i+1]);
        return edges;           /* create, store, and return */
      }                         /* the base of the triangle */
      n = this.verts.length;    /* get modulo for id computations */
      i = -i;                   /* and the tip position in the key */
      v = this.path[0];         /* get the triangle tip and */
      p = this.key[0];          /* the first and last key vertex */
      q = this.key[e = this.cnt-1];
      if (q == null) q = this.key[--e];
      if (((i <= 1) || (e < 2)) /* if triangle cuts off the anchor */
      &&  (v.id == (p.id +n -this.dir) % n)) {
        edges    = new Edge[1]; /* there is only one edge */
        edges[0] = new Edge(v, this.key[1]);
        return edges;           /* create, store, and return */
      }                         /* the right side of the triangle */
      if (((i >= e) || (e < 2)) /* if triangle cuts off the key end */
      &&  (v.id == (q.id +n +this.dir) % n)) {
        edges    = new Edge[1]; /* there is only one edge */
        edges[0] = new Edge(v, this.key[e-1]);
        return edges;           /* create, store, and return */
      }                         /* the left side of the triangle */
      if      (i <= 0) i++;     /* adapt extreme */
      else if (i >  e) i--;     /* key index values */
      edges    = new Edge[2];   /* if standard triangle in the middle */
      edges[0] = new Edge(v, this.key[i-1]);
      edges[1] = new Edge(v, this.key[i]);
      return edges;             /* create the two triangle sides */
    }                           /* that are adjacent to the tip */

    /* --- triangle with attached path --- */
    if (this.pdir != this.keydir) {
      q = this.key[1];          /* if path starts at start of key, */
      p = this.key[0]; }        /* get the first two key vertices */
    else {                      /* if path starts at end of key */
      p = this.key[e = this.cnt-1];
      if (p == null) p = this.key[--e];
      q = this.key[e-1];        /* get the last  two key vertices */
    }                           /* (get the base of the triangle) */
    edges = new Edge[i = this.plen+1];
    edges[--i] = new Edge(p, this.path[0]);
    edges[--i] = new Edge(q, this.path[0]);
    while (--i >= 0)            /* create and add triangle edges */
      edges[i] = new Edge(this.path[i], this.path[i+1]);
    return edges;               /* create and return the edges */
  }  /* edges() */              /* on the best splitting path */

}  /* class Combined2 */
