/*----------------------------------------------------------------------
  File    : PGView.java
  Contents: a simple pointgon viewer
  Author  : Christian Borgelt
  History : 2005.02.20 file created from file PGView.java
            2005.02.26 mouse listeners for panning and scaling added
            2005.03.03 random pointgon generation added
            2005.03.09 some menu and message extensions
            2005.03.19 adding and removing edges manually made possible
            2005.03.23 dialog box for minimum weight triangulation added
            2005.04.19 about dialog box background generalized
            2005.04.20 status text update through timer
            2005.05.12 extended to triangle-based algorithm
            2005.05.14 invokeLater changed to invokeAndWait
            2005.12.15 optional use of more precomputed info. added
            2006.01.19 adapted to third  subproblem splitter
            2006.02.08 adapted to fourth subproblem splitter
            2006.02.18 greedy triangulation added to list of algorithms
            2013.04.22 adapted to type argument of JComboBox
            2014.10.23 changed from LGPL license to MIT license
            2018.11.14 replaced some deprecated functions
----------------------------------------------------------------------*/
package pointgon;

import java.io.*;
import java.util.Random;
import java.text.DecimalFormat;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.imageio.ImageIO;

/*--------------------------------------------------------------------*/
/** Class for a logo panel.
 *  @author Christian Borgelt
 *  @since  2005.02.20 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
class LogoPanel extends JPanel {

  private static final long serialVersionUID = 0x00030006L;

  /** the data for the Otto logo (GIF format) */
  private byte[] otto = {
      71,  73,  70,  56,  57,  97, 100,   0, 101,   0, -25,   0,
       0,   0,   0,   0,   1,   1,   1,   2,   2,   2,   3,   3,
       3,   4,   4,   4,   5,   5,   5,   6,   6,   6,   7,   7,
       7,   8,   8,   8,   9,   9,   9,  10,  10,  10,  11,  11,
      11,  12,  12,  12,  13,  13,  13,  14,  14,  14,  15,  15,
      15,  16,  16,  16,  17,  17,  17,  18,  18,  18,  19,  19,
      19,  20,  20,  20,  21,  21,  21,  22,  22,  22,  23,  23,
      23,  24,  24,  24,  25,  25,  25,  26,  26,  26,  27,  27,
      27,  28,  28,  28,  29,  29,  29,  30,  30,  30,  31,  31,
      31,  32,  32,  32,  33,  33,  33,  34,  34,  34,  35,  35,
      35,  36,  36,  36,  37,  37,  37,  38,  38,  38,  39,  39,
      39,  40,  40,  40,  41,  41,  41,  42,  42,  42,  43,  43,
      43,  44,  44,  44,  45,  45,  45,  46,  46,  46,  47,  47,
      47,  48,  48,  48,  49,  49,  49,  50,  50,  50,  51,  51,
      51,  52,  52,  52,  53,  53,  53,  54,  54,  54,  55,  55,
      55,  56,  56,  56,  57,  57,  57,  58,  58,  58,  59,  59,
      59,  60,  60,  60,  61,  61,  61,  62,  62,  62,  63,  63,
      63,  64,  64,  64,  65,  65,  65,  66,  66,  66,  67,  67,
      67,  68,  68,  68,  69,  69,  69,  70,  70,  70,  71,  71,
      71,  72,  72,  72,  73,  73,  73,  74,  74,  74,  75,  75,
      75,  76,  76,  76,  77,  77,  77,  78,  78,  78,  79,  79,
      79,  80,  80,  80,  81,  81,  81,  82,  82,  82,  83,  83,
      83,  84,  84,  84,  85,  85,  85,  86,  86,  86,  87,  87,
      87,  88,  88,  88,  89,  89,  89,  90,  90,  90,  91,  91,
      91,  92,  92,  92,  93,  93,  93,  94,  94,  94,  95,  95,
      95,  96,  96,  96,  97,  97,  97,  98,  98,  98,  99,  99,
      99, 100, 100, 100, 101, 101, 101, 102, 102, 102, 103, 103,
     103, 104, 104, 104, 105, 105, 105, 106, 106, 106, 107, 107,
     107, 108, 108, 108, 109, 109, 109, 110, 110, 110, 111, 111,
     111, 112, 112, 112, 113, 113, 113, 114, 114, 114, 115, 115,
     115, 116, 116, 116, 117, 117, 117, 118, 118, 118, 119, 119,
     119, 120, 120, 120, 121, 121, 121, 122, 122, 122, 123, 123,
     123, 124, 124, 124, 125, 125, 125, 126, 126, 126, 127, 127,
     127,-128,-128,-128,-127,-127,-127,-126,-126,-126,-125,-125,
    -125,-124,-124,-124,-123,-123,-123,-122,-122,-122,-121,-121,
    -121,-120,-120,-120,-119,-119,-119,-118,-118,-118,-117,-117,
    -117,-116,-116,-116,-115,-115,-115,-114,-114,-114,-113,-113,
    -113,-112,-112,-112,-111,-111,-111,-110,-110,-110,-109,-109,
    -109,-108,-108,-108,-107,-107,-107,-106,-106,-106,-105,-105,
    -105,-104,-104,-104,-103,-103,-103,-102,-102,-102,-101,-101,
    -101,-100,-100,-100, -99, -99, -99, -98, -98, -98, -97, -97,
     -97, -96, -96, -96, -95, -95, -95, -94, -94, -94, -93, -93,
     -93, -92, -92, -92, -91, -91, -91, -90, -90, -90, -89, -89,
     -89, -88, -88, -88, -87, -87, -87, -86, -86, -86, -85, -85,
     -85, -84, -84, -84, -83, -83, -83, -82, -82, -82, -81, -81,
     -81, -80, -80, -80, -79, -79, -79, -78, -78, -78, -77, -77,
     -77, -76, -76, -76, -75, -75, -75, -74, -74, -74, -73, -73,
     -73, -72, -72, -72, -71, -71, -71, -70, -70, -70, -69, -69,
     -69, -68, -68, -68, -67, -67, -67, -66, -66, -66, -65, -65,
     -65, -64, -64, -64, -63, -63, -63, -62, -62, -62, -61, -61,
     -61, -60, -60, -60, -59, -59, -59, -58, -58, -58, -57, -57,
     -57, -56, -56, -56, -55, -55, -55, -54, -54, -54, -53, -53,
     -53, -52, -52, -52, -51, -51, -51, -50, -50, -50, -49, -49,
     -49, -48, -48, -48, -47, -47, -47, -46, -46, -46, -45, -45,
     -45, -44, -44, -44, -43, -43, -43, -42, -42, -42, -41, -41,
     -41, -40, -40, -40, -39, -39, -39, -38, -38, -38, -37, -37,
     -37, -36, -36, -36, -35, -35, -35, -34, -34, -34, -33, -33,
     -33, -32, -32, -32, -31, -31, -31, -30, -30, -30, -29, -29,
     -29, -28, -28, -28, -27, -27, -27, -26, -26, -26, -25, -25,
     -25, -24, -24, -24, -23, -23, -23, -22, -22, -22, -21, -21,
     -21, -20, -20, -20, -19, -19, -19, -18, -18, -18, -17, -17,
     -17, -16, -16, -16, -15, -15, -15, -14, -14, -14, -13, -13,
     -13, -12, -12, -12, -11, -11, -11, -10, -10, -10,  -9,  -9,
      -9,  -8,  -8,  -8,  -7,  -7,  -7,  -6,  -6,  -6,  -5,  -5,
      -5,  -4,  -4,  -4,  -3,  -3,  -3,  -2,  -2,  -2,  -1,  -1,
      -1,  33,  -7,   4,   1,  10,   0,  -1,   0,  44,   0,   0,
       0,   0, 100,   0, 101,   0,   0,   8,  -2,   0,  -1,   9,
      28,  72, -80, -96, -63,-125,   8,  15,   2,  88, -56, -80,
      33, -61,-124,  16,  35,  74,-100,  72,-111,  98, -61,-119,
      14,  43, 106, -36, -56,  49, -31, -61,-114,   2,  51,-126,
      28,  73, 114, -32,  71,-123,  33,  77, -86,  68, -72, -80,
     -92,  75,-115,  45,  81, -90,  -4,  23,  19, -64,  74,-101,
      10,  99, -66, -36,  73,  80,  39, -53,-101,  56, 113, -46,
    -100,  41, -76, -96,  79, -98,  36,-113,  30,  29,  74, 116,
     104, -48,-101,  33,-117,  70,  69,-102, -44, -96, -44, -89,
      77, 105,  98, 117, -86,  82, -22,  84, -86,  48,-123,-118,
     -43, 106,  85, -25,  88, -90, 100, -65, -86, -99,   9,  54,
     -94, -46, -98,  47,-117, -38,  28, 123, -43, 107, -37,-107,
      89,  83,  46,   5, -23, 115,  43, -36, -69,  70, -83, 106,
      61,-117,  84, -20,  73, -71,-128, -43,-102, -11, 123, 119,
     -21, -36, -65, 109,  17,   7,  78,  44,-109, 107,  79, -69,
      37, -81,  10, -90,  44,  -8,  49,  94, -76, 113,  55,-125,
     -26,-116, -46, -16, 100,-105,-110,-103,  98,  38,-115,-105,
     -16, -24,-114, 117,  77, -82,  14, 123,-111, -17, -23, -39,
      24, 127, -14, 117, -56,  91,  36,  76, -93, -82,  57,-102,
      -2, 108, -79, -73, 113, -34, -74, -47, -30, -10,   8,-107,
      45, -58, -29, -48, 125,  87, -60, -68,  92, -76, 106, -31,
     -47, -77,  35, -49, -67, -71,  58, 113, -78, 123,  89,  -2,
      70,  31, -84,  -3, -28, 116, -84, -34,-107, 127, 119,-101,
      61,  42, 114, -24, -65,  65, 123,  63, -53,  24, -94, 118,
     -32,  25,-113,  79, -73, -68,  94, -76, 103, -38, -58,  61,
      55,  87,-128,  97, -27, -59, -36,  95, -23,-111,  87,  91,
     124,   4, -18, -57, -42, 108, 118,  -3,  39,-111, 126,  27,
     -43, -44,-101,-125, -51, -23,  38,  31,-128, -46, -99,  39,
     -37,-126,  21,   6,   7,  89,  90,  33,  54, -72,  27,-120,
     -59,  25,  56,  98, 127,  30, -63,-105, -44, 118,  22,  65,
      22, -95, 115,   9,  42,-120, -30,-119,  29, -62, -26, -97,
    -120, -10,  81,  -8, -30,-123,-103,  89,  55,  88,-119,  64,
      -2,   8,  99,  85,  43, -98, 104,  35, 106,  62,-114,-108,
      90, 125, -20,  61,-108,  35, 108,  77, -98,  71,  24, 122,
     -81,-119,-105, -97, 121,  70,  30,  -7, -36,-125, -51,  73,
     -88, -27,-106,  38,  18, -23,  37,-122, -22, 109, -24, -42,
    -112, -18,  85, 121,  89,  75,  74, -71,  -7, -91,-127,  80,
      86,  86, -34, 106,  -7, -31,  87, 100,-120,  93,   1,  87,
     -36, 120, 111,  34,  88,-103,-115,  83, -74, -88,  89,-115,
     119,  14, 104, -95,  92, 120, -66,  89, -90, 112,  89, -10,
    -104, -24,  71,  92, -74,  73, -36, 125, -69,  97,-105,  40,
    -101,  17,-118,  89,  87, 123,-110,  50, 106, 102, 121,  75,
    -106, 117, -29,-121,-128, -82, 105,  33, -86, 120, -34, -23,
     -36, -85,  -2, 105,  77,  73, -22,-124, -79, 122,   5, -95,
     -85, -18, -11,  55,  96, -92, -77,  26,  42, 106,-110, -90,
    -126,  90, -42,-124, -99,  98, -54,  39,-117,-114, -70,-104,
    -109,-109, -67,  22, -40,  38, 110, -97, -98, -70, -42, -88,
     -54,-114,  41,  91,  92, 103,   2,  -5, -89, -79,-125,   6,
      85,  99,  78, -39, -62, -54,  33, -73,  35, 122, -21, -24,
     -74, -31,  70, 106,  37, -82,-127, -95, -25,  90, 117, -87,
      94,  38, 100,-108, -20,  -6,  57,-100,-102, -95, -54, 105,
     -21, -72, -62, -46, -22, -41, 114, -28, 126,   5,  39, -75,
     -43, -86,-118,  42, -70, 123, -54,  -5,  45, -95,   5, -77,
     -38,  39,-114,-113,  90,  10,  41,-100,  17, 127,  70,  23,
     -77,   5,   3, -71, -16, -77, -31,-107,  27, -90, -76,-116,
      86,  27, 112, -69,  12,  87,  42, -45,-107, -13,  10, -36,
     -98, -73,-108, -90, -53, 106, -59, -61, -74, 118,  26, -72,
      -9, -99,  41, -27,  94,  44,-113,  44, -18, -69, -12,  78,
     -86, -24, -51,   3, 119, -90,  95, -72, -69, 122, -84, 110,
     -78,-109,-106, 102, -14, -49, -17,  73, -85,  90,  77,  51,
      35,  92, -90,-108, -90,  42,  77,  33,-124,  89, 105, -90,
     105, -68, 126, -82,-101,  48, 119, -81, 125, -37, -20, -76,
     -86,  38, -35, -75,-118, -81,-102, -84, -25, -93,  29, 119,
      23, -32, -41,  78, 105,  61,  51,-113,  72,  39, -36, -74,
     -68,  37,  75, -36,  93,  -2, -84,   7, -90,  70, 115, -58,
     -16,  26,  38, -89, -98, 104, 103, 120, -12, -40, 119,-109,
      88, -78, -45, 112, -43,  25, -90, -92,-117,  47, 122,  93,
     -52,  68, 113, -67,  22, -35,   8,   2,-100, -13,-123,  23,
      29, -39, -41, -26,  45, -17,  29, -75,-122, -13,  10,  62,
      52, -59, -28, -79, -39,  26,-116,  75,   5,  55, -33, -96,
     -63,  -6, -36,  97,-100,-116, -29,-101,  30, 125,  61, -53,
     110,  30,-105,  44,  -9, -99, -78, 125, -41,  18,  43, 120,
     -28, -66, -31,-100, -77, -99, 120, -57,  23,  54, -27,-107,
      99, -86, -74, 124, -44, -55, -84,  35,-115, -30,  89, -68,
     -14, -35,  29, -33,  60,  58,-102,-114,  39,  95, 119, -19,
     -34, -97,-116,  44, -40,-118,   3, -49,  28,  -8, -80,-109,
      -8, -33, -58, 115,   7, -70, 125, -29,  47,  10,  10, -26,
      78,-110, -67, 123, -21, -47,  96,  -5,-115,  63, -92,  70,
    -117, -21, 126,-112,-100, -46,  22,   0, -87, -25, -65, -58,
     -79, -17,  59,  -6,  43,  12, -18, -56, -42, -75, -60, -11,
     -56, 123,   7,-116, -47, -27, -40,  19,  62, -27, -91, -81,
      49,  23, -84, -96,   3, 127,  87,  62, -50, -56,-115, 121,
      28, -84,  30,   8,  89, -93, -66,   3, -75, -56,-126, -63,
      35, -31, -80,  58,  53,  39,  98, -99,  77,-123,  34, -84,
     -32,   9,  41,-104,  66,  24,-102, -81,-122,  45, -44, -48,
       6, 109,  -8, -85,  99,  45, 107,-121,  44,  54,  84, -50,
     -13,  30,  24, -84,  32,  86,  72,  49,  67,  28,  97,   4,
      85,  40, -90, -17, -59,  11,-120,  70, -60,  97, 102,  48,
      23, -59,  27, -14,   4, 117,  85, -52,  98, -82, -76, -56,
     -59,  37,  82,  36,  32,   0,  59 };

  /** the data for the ECSC logo (GIF format) */
  private byte[] ecsc = {
      71,  73,  70,  56,  55,  97, 101,   0, 100,   0, -91,  38,
       0,   0,-105, -40,   2,-104, -40,   4,-104, -40,   6,-103,
     -39,  11,-101, -39,  21, -97, -38,  27, -95, -37,  35, -92,
     -37,  56, -84, -35,  66, -81, -34,  69, -79, -33,  87, -73,
     -32,  89, -72, -31,  92, -71, -31, 105, -66, -30, 114, -63,
     -29, 121, -60, -28,-127, -57, -28, -99, -47, -25, -97, -46,
     -25, -94, -45, -24, -91, -44, -24, -89, -44, -24, -69, -36,
     -22, -66, -35, -22, -63, -34, -21, -53, -30, -20, -50, -29,
     -20, -46, -28, -20, -45, -28, -21, -42, -27, -19, -41, -26,
     -19, -41, -25, -19, -39, -25, -19, -34, -23, -19, -29, -20,
     -17, -24, -19, -18, -17, -17, -17,  -1,  -1,  -1,  -1,  -1,
      -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
      -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
      -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
      -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
      -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
      -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
      -1,  44,   0,   0,   0,   0, 101,   0, 100,   0,   0,   6,
      -2, -64,-110, 112,  72,  44,  26,-113, -56, -92,  16,  84,
      89,  24,   4,-122,  69,   5, -92, -84,  90, -81, -40, 108,
     -74,   3,   1, 120, -65,  95,  72,  71,  75,  46,-101, -55,
    -101,   1, 120, -19,  29, 108, -50, -16, 120, 124, -61, -82,
     123, -33, -14, -68,  94, -39,  81, -37, -39,   3,  99, 123,
    -125,-124,  93, 127, 117,  16,-124,-118, 121,  32,-121, 127,
      84,-117,-111, 101,  21,-114, 118,  21,-110,-104,  89,  11,
    -107, 117,  11,-103, -97,  85,   6,-100, 108,   6, -96, -90,
      71,   2, -93, 107,   2, -89, -83,  67, -94, -86,  95, -91,
     -82, -83,-101, -79,  94, -98, -76, -89,-108, -73,   0,-105,
     -70, -90,-115, -67,-112, -64, -96,-122, -86,-119, -59, -90,
     125, -86,-127, -54, -89, 116, -93, 120, -49, -90, 105,-114,
     110,-103,  28,  14,   5,   5,  14,  28, -44,  92, 127,  98,
    -103,  22, 108,  22, -44,  37,  76,  78,  80,  82, -60,-110,
      28, 118, -33, -24, -70,  14, 118,  14, -13, -70,   5, 118,
       5,  -8, -76,  -6, 117,  -4,  -6, -75, -86,  87, -25, -98,
     -64,  43,  35,  50,  68,  72,  80,  64,  64,-127,   4,  17,
      50,-116, -64,   2, -81,-114, -68,-125,  73,  60,  68, 112,
      20, -63, -61,-107, 114, 107, -50,  97,  76,  50,  65, -43,
    -124,  43, -38, -72, 121,  27,-119, 100,  68,-125,  91,  13,
      38, -78,-116,  51,  34,  65,  47,   0,   9, 100, -50,  52,
     -13, -14,  38,  -2,-128,   6,  59, -51,-108, -12, -23, -27,
     100, -48,  44,  30,-120,-126, -15, 120, -12, -54,  70, -91,
      94,  34,  52, -75,  50,   2,  42,  24, -99,  83,-113, 100,
     -80,  -6,  37,  67, -42,  36,  79, -71,  74,  -3, 122, -60,
      38,  87,-100, 100,-113,  -4, -29,  26,  48,  45,-111,  84,
     103,  89, -71,  37, -78, -42, 106, -37,  76,  36,  52,  80,
     120, -96, -32,   0,-127,   0,   4,  14,  40, 120,  64,  65,
       3,-119,  51, 102, -71,  38, -56,  36,  66,   2,   2,  85,
       8,  36,-120,  32,  19, -42, -22, -40,  69,  24,  20,  16,
      85,-128,  33, -53, -42, -77,  94,  21,  93,  56, -64, -11,
     -64,   5,-124, 103,   1,  96, -51, -13,  65, 115, 106,   0,
      10,  62,  88, -87,  76, -12, -78,  30,   9, -81, -41,  72,
     -88,-110, -44,  42, -45,  60,  36,  24, -28, 102, -61, -32,
      48,  73, -88,  70, -13,-120,  32,  61,-100, -51,-127, -55,
      73, 122, -34,   4, -86,  39,   4, -36, -26, 108,   4,-124,
      72,  82, -13, 102,  78,  61,  34, -82,  99, -49,  14,  -3,
    -120,  75,-104, -85, -49,-112,  96,  62, -34, -50,   1, -29,
      72,-122, 114,  74, -98,  71, 120,  -5,  67,  12, 120, -45,
     102, -45, 113,  16, -18,  -5,-114, -20,  86,  69,  66,  11,
      53, -12,  80,  68, -23, -59, -15,   1,-128,-100, -56,  54,
    -113, 107,  12,  30, -94, -64,  60,  23,  68, -56, -55, 105,
     -44, -80, 103, -31,  31,   7,  -2,  80,-125, -63,-122,-100,
     116, -90,  12,-124,  32,  -2,  49,  97,  49,  34,-108, -56,
      73, 121, 123, -92, -44, -51,  69,  70,  -4, -89, -30,  33,
       2, -18,   1,  18,  24,  34,  25, -15, -40,-116,-121,  32,
      48,  72,  69, 108, -64,  40,   4,   9,  60,  86,   2, -97,
      28,   4, -79,  97,  16,  17,  26,  20, -23,-120,   6, 123,
     -44,   5, -58,  93,  66,  80, -32, -28,  33,  20,  68, -71,
    -113,  17,  15,  92,  -7, -57,   3, 123,  36, -71, -58,-110,
      67,-112, -24,  37,  24,  39, -26,   1, -28,  26,  66,-106,
     -96, -31,-103,  95, 116, 104, -93,  57,  71,  16,   0,  39,
      27,   4,  16, -30, -30,  74,  71,   4, 112, -25,  26,   1,
     -24,  98, -25, -97,  95, -28,  73, -53,-101, 119, -54, -23,
    -118,-103, 119, -90, -39,  74,-105,-124, 122,   1,  38,  45,
      86,  70,  10,  64,-106, -76,  52, 105,  41,-108, -76,  16,
     105, -23,-111, -83, -20,  -8, -89,-113, -64, -56, 120, 103,
    -115, -76, -92,  72,  40,-117, -76,  48, -22, -92, -93, -70,
     124, 120, -89,-120, -54,  32, -54, -93, -94, -54,  84, 120,
      38,-122, -44, -72,  90,  34, -84, -54,  44, 120, -91,-125,
     -13,-104, -86,  34, -86, -13, -40, -89,  98, 126,   7, -83,
     -89, -30, 123,  35,-123,   7, -94,   0, -84,  10, 100, -99,
    -123, -38,   5, -75,  28,-125, -49,  53,  21, -36, 125, -59,
     125, 101, -20, 107, -56,  78, -43,  90, 110,  38, -79, -51,
      37, -60, 104, -91, -15, -86, -82,  16,-103, 109,  70, -21,
     -69,  69,  52,  38, 106,  37,-111,  85,  75, -81,  17, 121,
     -19, -43, -41,  95,-127,  13,  86,  24, -88,  65,   5,   1,
      0,  59 };

  /** the logo as an image icon */
  private ImageIcon logo = new ImageIcon(ecsc);

  /*------------------------------------------------------------------*/
  /** Get the preferred panel size.
   *  @return the preferred panel size
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Dimension getPreferredSize ()
  { return new Dimension(logo.getIconWidth(), logo.getIconHeight()); }

  /*------------------------------------------------------------------*/
  /** Draw the panel content
   *  @param  g     the graphics context
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/
  public void paint (Graphics g)
  { g.drawImage(logo.getImage(), 0, 0, this); }

}  /* class LogoPanel */


/*--------------------------------------------------------------------*/
/** Class for a pointgon viewer window.
 *  @author Christian Borgelt
 *  @since  2005.02.20 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public class PGView extends JFrame
  implements MouseListener, MouseMotionListener, Runnable {

  private static final long serialVersionUID = 0x00030009L;
  /** the version string */
  public  static final String VERSION = "3.9 (2018.11.14)";

  /** the font for input fields */
  private static final Font font  = new Font("Dialog", Font.BOLD,  12);
  /** the font for help texts */
  private static final Font small = new Font("Dialog", Font.PLAIN, 10);

  /** whether run as a program */
  private boolean      isprog  = false;
  /** special menu items */
  private JMenuItem[]  mitems  = null;
  /** scroll pane viewport */
  private JScrollPane  scroll  = null;
  /** pointgon viewer panel */
  private PGPanel      panel   = null;
  /** status bar for messages */
  private JTextField   stat    = null;
  /** pointgon generation dialog box */
  private JDialog      randpg  = null;
  /** "MWT..." dialog box */
  private JDialog      runmwt  = null;
  /** "About..." dialog box */
  private JDialog      about   = null;
  /** a file chooser */
  private JFileChooser chooser = null;
  /** current pointgon file */
  private File         curr    = null;
  /** current pointgon */
  private Pointgon     pgon    = null;
  /** flag for state of pointgon */
  private boolean      pgvalid = false;
  /** minimum weight triangulation solver */
  private MWT          mwt     = null;
  /** minimum weight triangulation solver mode */
  private int          mwtmode = MWT.COMB_2|MWT.PRECHECK|MWT.REVKEY;
  /** thread for computations */
  private transient Thread thread = null;
  /** timer for status update */
  private Timer        timer   = null;
  /** flag for running minimum weight triangulation computation */
  private boolean      running = false;
  /** flag for interruption */
  private boolean      stopped = false;
  /** vertex for edge insertion */
  private Vertex       vertex  = null;
  /** mouse operation mode */
  private int          mode    = 0;
  /** x-coordinate of mouse position */
  private int          mx;
  /** y-coordinate of mouse position */
  private int          my;
  /** buffer for scaling factor */
  private double       scale;
  /** buffer for scaling factor */
  private double       factor;

  /*------------------------------------------------------------------*/
  /** Create a file chooser.
   *  @return the created file chooser
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private JFileChooser createChooser ()
  {                             /* --- create a file chooser */
    JFileChooser fc = new JFileChooser();
    fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
    fc.setFileHidingEnabled(true);
    fc.setAcceptAllFileFilterUsed(true);
    fc.setMultiSelectionEnabled(false);
    fc.setFileView(null);       /* create a standard file chooser */
    return fc;                  /* without customized filters */
  }  /* createChooser() */

  /*------------------------------------------------------------------*/
  /** Load the current pointgon from a file.
   *  @param  file  the file to load the pointgon from
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void loadPointgon (File file)
  {                             /* --- load a pointgon */
    if (file != null)           /* if a file name is given, */
      this.curr = file;         /* set it as the current file, */
    else {                      /* otherwise show file chooser */
      if (this.chooser == null) this.chooser = this.createChooser();
      this.chooser.setDialogType(JFileChooser.OPEN_DIALOG);
      int r = this.chooser.showDialog(this, null);
      if (r != JFileChooser.APPROVE_OPTION) return;
      this.curr = this.chooser.getSelectedFile();
    }                           /* get the selected file */
    try {                       /* load the pointgon */
      this.pgon = new Pointgon(new FileInputStream(this.curr));
      this.pgvalid = this.pgon.isValid(); /* check the pointgon */
      this.panel.setPointgon(this.pgon);  /* and then store it */
      this.stat.setText((this.pgvalid         ? "" : "in")   +"valid "
                       +(this.pgon.isConvex() ? "" : "non-") +"convex "
                       +"pointgon loaded ("+this.curr.getName()+")."); }
    catch (IOException e) {     /* catch and report i/o errors */
      String msg = "file " +this.curr.getName() +":\n" +e.getMessage();
      this.stat.setText(msg); System.err.println(msg);
      JOptionPane.showMessageDialog(this, msg,
        "Error", JOptionPane.ERROR_MESSAGE);
    }                           /* set the status text */
  }  /* loadPointgon() */

  /*------------------------------------------------------------------*/
  /** Save the current pointgon to a file.
   *  @param  file  the file to save the pointgon to
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void savePointgon (File file)
  {                             /* --- save a pointgon */
    if (file != null)           /* if a file name is given, */
      this.curr = file;         /* set it as the current file, */
    else {                      /* otherwise show file chooser */
      if (chooser == null) chooser = createChooser();
      chooser.setDialogType(JFileChooser.SAVE_DIALOG);
      int r = chooser.showDialog(this, null);
      if (r != JFileChooser.APPROVE_OPTION) return;
      this.curr = chooser.getSelectedFile();
    }                           /* get the selected file */
    try {                       /* create a writer for the file */
      FileWriter writer = new FileWriter(this.curr);
      writer.write(this.pgon.toString());
      writer.close(); }         /* save the current pointgon */
    catch (IOException e) {     /* catch and report i/o errors */
      String msg = "file " +this.curr.getName() +":\n" +e.getMessage();
      this.stat.setText(msg); System.err.println(msg);
      JOptionPane.showMessageDialog(this, msg,
        "Error", JOptionPane.ERROR_MESSAGE);
    }                           /* set the status text */
  }  /* savePointgon() */

  /*------------------------------------------------------------------*/
  /** Save the Window contents as an image.
   *  @param  file  the file to save the image to
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void saveImage (File file)
  {                             /* --- save image to a file */
    if (file == null) {         /* if no file name is given */
      if (this.chooser == null) this.chooser = this.createChooser();
      this.chooser.setDialogType(JFileChooser.SAVE_DIALOG);
      int r = this.chooser.showDialog(this, null);
      if (r != JFileChooser.APPROVE_OPTION) return;
      file = this.chooser.getSelectedFile();
    }                           /* let the user choose a file name */
    try {                       /* open an output stream */
      FileOutputStream stream = new FileOutputStream(file);
      ImageIO.write(this.panel.makeImage(), "png", stream);
      stream.close(); }         /* save the decision tree image */
    catch (IOException e) {     /* catch and report i/o errors */
      String msg = e.getMessage();
      stat.setText(msg); System.err.println(msg);
      JOptionPane.showMessageDialog(this, msg,
        "Error", JOptionPane.ERROR_MESSAGE);
    }                           /* set the status text */
  }  /* saveImage() */

  /*------------------------------------------------------------------*/
  /** Generate a random pointgon.
   *  @param  vertcnt  the number of perimeter vertices
   *  @param  holecnt  the number of holes
   *  @param  radius   the radius parameter
   *                   (maximum perimeter vertex distance from origin)
   *  @param  seed     the seed value of the random number generator
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void genPointgon (int vertcnt, int holecnt,
                            double radius, long seed)
  {                             /* --- generate random pointgon */
    Random rand = (seed > 0) ? new Random(seed) : new Random();
    this.pgon = new Pointgon(vertcnt, holecnt, radius, rand);
    this.pgvalid = true;        /* generate a random pointgon */
    this.pgon.transform(12.0, 0, 0);
    this.panel.setPointgon(PGView.this.pgon);
    this.repaint();             /* repaint the window contents */
    this.stat.setText("random pointgon generated.");
    this.curr = null;           /* invalidate the file name */
  }  /* genPointgon() */

  /*------------------------------------------------------------------*/
  /** Create the random pointgon creation dialog box.
   *  @return the created dialog box
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private JDialog createRandPG ()
  {                             /* --- create "Generate..." dialog */
    final JDialog      dlg  = new JDialog(this,
                                    "Generate Random Pointgon...");
    GridBagLayout      g    = new GridBagLayout();
    GridBagConstraints lc   = new GridBagConstraints();
    GridBagConstraints rc   = new GridBagConstraints();
    JPanel             grid = new JPanel(g);
    JPanel             bbar;
    JLabel             lbl;
    JTextArea          help;
    JButton            btn;

    grid.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
    lc.fill      =              /* fill fields in both directions */
    rc.fill      = GridBagConstraints.BOTH;
    rc.weightx   = 1.0;         /* resize only the input fields, */
    lc.weightx   = 0.0;         /* but not the labels */
    lc.weighty   = 0.0;         /* resize lines equally */
    rc.weighty   = 0.0;         /* in vertical direction */
    lc.ipadx     = 10;          /* gap between labels and inputs */
    lc.ipady     = 10;          /* make all lines of the same height */
    rc.gridwidth = GridBagConstraints.REMAINDER;

    lbl = new JLabel("Perimeter vertices:");
    g.setConstraints(lbl, lc);     grid.add(lbl);
    final JSpinner vertcnt = new JSpinner(
          new SpinnerNumberModel(10, 3, 999999, 1));
    g.setConstraints(vertcnt, rc); grid.add(vertcnt);

    lbl = new JLabel("Inner/hole vertices:");
    g.setConstraints(lbl, lc);     grid.add(lbl);
    final JSpinner holecnt = new JSpinner(
          new SpinnerNumberModel(6, 0, 999999, 1));
    g.setConstraints(holecnt, rc); grid.add(holecnt);

    lbl = new JLabel("Inner radius (percent):");
    g.setConstraints(lbl, lc);    grid.add(lbl);
    final JSpinner radius = new JSpinner(
          new SpinnerNumberModel(20, 0, 100, 1));
    g.setConstraints(radius, rc); grid.add(radius);

    help = new JTextArea(
       "The radii for the perimeter vertices are generated\n"
      +"between the inner radius percentage and 100% (*12).");
    help.setFont(small); help.setEditable(false);
    help.setBackground(this.getBackground());
    g.setConstraints(help, rc); grid.add(help);

    lbl = new JLabel("Seed for random numbers:");
    g.setConstraints(lbl, lc);    grid.add(lbl);
    final JSpinner seed = new JSpinner(
          new SpinnerNumberModel(0, 0, 999999, 1));
    g.setConstraints(seed, rc); grid.add(seed);

    help = new JTextArea(
       "If the seed for the pseudo-random number generator\n"
      +"is set to zero, the system time will be used instead.");
    help.setFont(small); help.setEditable(false);
    help.setBackground(this.getBackground());
    g.setConstraints(help, rc); grid.add(help);

    bbar = new JPanel(new GridLayout(1, 2, 5, 5));
    bbar.setBorder(BorderFactory.createEmptyBorder(0, 5, 5, 3));
    btn = new JButton("Ok"); bbar.add(btn);
    btn.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        dlg.setVisible(false);
        PGView.this.genPointgon(
          ((Integer)vertcnt.getValue()).intValue(),
          ((Integer)holecnt.getValue()).intValue(),
          ((Integer)radius.getValue()).intValue() /100.0,
          ((Integer)seed.getValue()).longValue());
      } } );
    btn = new JButton("Apply"); bbar.add(btn);
    btn.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        PGView.this.genPointgon(
          ((Integer)vertcnt.getValue()).intValue(),
          ((Integer)holecnt.getValue()).intValue(),
          ((Integer)radius.getValue()).intValue() /100.0,
          ((Integer)seed.getValue()).longValue());
      } } );
    btn = new JButton("Close"); bbar.add(btn);
    btn.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        dlg.setVisible(false); } } );

    dlg.getContentPane().add(grid, BorderLayout.CENTER);
    dlg.getContentPane().add(bbar, BorderLayout.SOUTH);
    dlg.setLocationRelativeTo(this);
    dlg.pack();
    return dlg;
  }  /* createRandPG() */

  /*------------------------------------------------------------------*/
  /** Create the run minimum weight traingulation dialog box.
   *  @return the created dialog box
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private JDialog createRunMWT ()
  {                             /* --- create "MWT..." dialog */
    final JDialog      dlg  = new JDialog(this,
                               "Minimum Weight Triangulation...", true);
    GridBagLayout      g    = new GridBagLayout();
    GridBagConstraints lc   = new GridBagConstraints();
    GridBagConstraints rc   = new GridBagConstraints();
    JPanel             grid = new JPanel(g);
    JPanel             bbar;
    JLabel             lbl;
    JTextArea          help;
    JButton            btn;
    int                i;

    grid.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
    lc.fill      =              /* fill fields in both directions */
    rc.fill      = GridBagConstraints.BOTH;
    rc.weightx   = 1.0;         /* resize only the input fields, */
    lc.weightx   = 0.0;         /* but not the labels */
    lc.weighty   = 0.0;         /* resize lines equally */
    rc.weighty   = 0.0;         /* in vertical direction */
    lc.ipadx     =  8;          /* gap between labels and inputs */
    lc.ipady     = 10;          /* make all lines of the same height */
    rc.gridwidth = GridBagConstraints.REMAINDER;

    lbl = new JLabel("MWT Algorithm:");
    g.setConstraints(lbl, lc);    grid.add(lbl);
    final JComboBox<String> algocb = new JComboBox<String>(new String[]{
      "triangle-based", "path-based", "combined 1", "combined 2",
      "greedy" });
    switch (this.mwtmode & MWT.ALGOMASK) {
      case MWT.TRIANGLE: algocb.setSelectedIndex(0); break;
      case MWT.PATH:     algocb.setSelectedIndex(1); break;
      case MWT.COMB_1:   algocb.setSelectedIndex(2); break;
      case MWT.COMB_2:   algocb.setSelectedIndex(3); break;
      default:           algocb.setSelectedIndex(4); break;
    }
    g.setConstraints(algocb, rc); grid.add(algocb);

    help = new JTextArea(
       "Whether a pointgon is split into subproblems\n"
      +"with paths through holes or with triangles.\n"
      +"The combined algorithms also use splits where\n"
      +"paths are attached to the tips of triangles.\n"
      +"Note that the greedy algorithm is a heuristic.");
    help.setFont(small); help.setEditable(false);
    help.setBackground(this.getBackground());
    g.setConstraints(help, rc); grid.add(help);

    lbl = new JLabel("Reverse keys:");
    g.setConstraints(lbl, lc);     grid.add(lbl);
    final JCheckBox revkey = new JCheckBox(
      "", (this.mwtmode & MWT.REVKEY) != 0);
    g.setConstraints(revkey, rc);  grid.add(revkey);

    help = new JTextArea(
       "Reversing the subproblem keys speeds up the\n"
      +"triangle-based and the combined 2 algorithm\n"
      +"and saves memory. For the path-based and the\n"
      +"combined 1 algorithm forward keys are better.");
    help.setFont(small); help.setEditable(false);
    help.setBackground(this.getBackground());
    g.setConstraints(help, rc); grid.add(help);

    lbl = new JLabel("Precheck triangles:");
    g.setConstraints(lbl, lc);     grid.add(lbl);
    final JCheckBox pretcb = new JCheckBox(
      "", (this.mwtmode & MWT.PRECHECK) != 0);
    g.setConstraints(pretcb, rc);  grid.add(pretcb);

    help = new JTextArea(
       "Precomputing and storing the valid triangles\n"
      +"can speed up some checks in the algorithms.");
    help.setFont(small); help.setEditable(false);
    help.setBackground(this.getBackground());
    g.setConstraints(help, rc); grid.add(help);

    lbl = new JLabel("Distribute holes:");
    g.setConstraints(lbl, lc);     grid.add(lbl);
    final JCheckBox distcb = new JCheckBox(
      "", (this.mwtmode & MWT.DISTHOLES) != 0);
    g.setConstraints(distcb, rc);  grid.add(distcb);

    help = new JTextArea(
       "Whether to check which subproblem a hole is in\n"
      +"and to assign it only to the containing one.\n"
      +"(This can be advantageous in connection with the\n"
      +"path-based algorithms as it reduces the number\n"
      +"of potential splits/paths that are generated.)");
    help.setFont(small); help.setEditable(false);
    help.setBackground(this.getBackground());
    g.setConstraints(help, rc); grid.add(help);

    bbar = new JPanel(new GridLayout(1, 2, 5, 5));
    bbar.setBorder(BorderFactory.createEmptyBorder(0, 5, 5, 3));
    btn = new JButton("Start"); bbar.add(btn);
    btn.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        dlg.setVisible(false);
        int i = algocb.getSelectedIndex();
        if      (i == 0) i = MWT.TRIANGLE;
        else if (i == 1) i = MWT.PATH;
        else if (i == 2) i = MWT.COMB_1;
        else if (i == 3) i = MWT.COMB_2;
        else             i = MWT.GREEDY;
        PGView.this.mwtmode
          = (i | ( revkey.isSelected() ? MWT.REVKEY    : 0)
               | ( pretcb.isSelected() ? MWT.PRECHECK  : 0)
               | ( distcb.isSelected() ? MWT.DISTHOLES : 0));
      } } );
    btn = new JButton("Cancel"); bbar.add(btn);
    btn.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        dlg.setVisible(false);
        PGView.this.mwtmode = -1;
    } } );

    dlg.getContentPane().add(grid, BorderLayout.CENTER);
    dlg.getContentPane().add(bbar, BorderLayout.SOUTH);
    dlg.setLocationRelativeTo(this);
    dlg.pack();
    return dlg;
  }  /* createRunMWT() */

  /*------------------------------------------------------------------*/
  /** Create the about dialog box.
   *  @return the created dialog box
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private JDialog createAbout ()
  {                             /* --- create "About..." dialog box */
    final JDialog dlg  = new JDialog(this, "About PGView...", true);
    Container     pane = dlg.getContentPane();
    LogoPanel     logo = new LogoPanel();
    JButton       btn  = new JButton("Ok");
    JPanel        rest = new JPanel(new BorderLayout(2, 2));
    JTextArea     text = new JTextArea
      ("PGView\n"
      +"A Pointgon Viewer including\n"
      +"Minimum Weight Triangulation\n"
      +"Version " +PGView.VERSION +"\n\n"
      +"European Centre for Soft Computing\n"
      +"c/ Gonzalo Gutierrez Quiros s/n\n"
      +"33600 Mieres, Asturias, Spain\n"
      +"christian@borgelt.net");
    text.setBackground(this.getBackground());
    text.setFont(new Font("Dialog", Font.BOLD, 12));
    text.setEditable(false);
    btn.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        dlg.setVisible(false); } } );
    rest.add(logo, BorderLayout.NORTH);
    rest.add(btn,  BorderLayout.SOUTH);
    pane.setLayout(new FlowLayout());
    pane.add(text);
    pane.add(rest);
    dlg.setLocationRelativeTo(this);
    dlg.pack();
    dlg.setResizable(false);
    return dlg;
  }  /* createAbout() */

  /*------------------------------------------------------------------*/
  /** Set the menu state.
   *  @param  running  whether a minimum weight triangulation is running
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void setMenuState (boolean running)
  {                             /* --- set the menu state */
    int i;                      /* loop variable */

    if (running) {              /* if running, disable menu items */
      if (this.randpg != null) this.randpg.setVisible(false);
      for (i = this.mitems.length; --i > 0; )
        this.mitems[i].setEnabled(false);
      this.mitems[0].setText("Abort Computations"); }
    else {                      /* if not running anymore */
      for (i = this.mitems.length; --i > 0; )
        this.mitems[i].setEnabled(true);
      this.mitems[0].setText("Minimum Weight Triangulation...");
    }                           /* enable menu items again */
  }  /* setMenuState() */

  /*------------------------------------------------------------------*/
  /** Update the status text.
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void updateStatus ()
  {                             /* --- update the status text */
    if (!this.running) return;  /* check for a running search */
    this.stat.setText(" running search ... (status: "
                     +this.mwt.getTriangles() +"/"
                     +this.mwt.getSubprobs()  +"/"
                     +this.mwt.getSplits()    +")");
  }  /* updateStatus() */       /* display search statistics */

  /*------------------------------------------------------------------*/
  /** Show the result of the computation.
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void showResult ()
  {                             /* --- show result of computation */
    long   time;                /* search time in milliseconds */
    double wgt;                 /* weight of triangulation result */
    Edge[] edges;               /* edges  of triangulation result */
    int    n;                   /* number of edges */
    String s;                   /* buffer for output */

    this.setMenuState(this.running = false);
    if (this.stopped) {         /* if execution has been aborted */
      this.stat.setText("minimum weight triangulation aborted.");
      JOptionPane.showMessageDialog(this,
        "Minimum weight triangulation aborted.",
        "Information", JOptionPane.INFORMATION_MESSAGE);
      return;                    /* show an abortion dialog */
    }                            /* and abort the function */
    wgt = this.mwt.getWeight();  /* get triangulation weight */
    if (wgt < 0) {               /* if something went wrong */
      this.stat.setText("minimum weight triangulation failed.");
      JOptionPane.showMessageDialog(this,
        "Something went wrong during the search,\n"
        +"maybe the program ran out of memory.\n"
        +"If this is not the case (check terminal),\n"
        +"please save the pointgon you processed\n"
        +"and send it to the author of this program.",
        "Alert", JOptionPane.ERROR_MESSAGE);
      return;                   /* show an alert and abort */
    }
    time  = this.mwt.getTime(); /* get more result information */
    edges = this.mwt.getEdges();/* (execution time and edges) */
    n = (edges != null) ? edges.length : 0;
    this.pgon.setEdges(edges);  /* set the edges in the pointgon */
    switch (this.mwtmode & MWT.ALGOMASK) {
      case MWT.TRIANGLE: s = "triangle"; break;
      case MWT.PATH:     s = "path";     break;
      case MWT.COMB_1:   s = "comb_1";   break;
      case MWT.COMB_2:   s = "comb_2";   break;
      default:           s = "greedy";   break;
    }                           /* get the name of the algorithm */
    s += (((this.mwtmode & MWT.PRECHECK)  != 0) ? ", precheck"    : "")
      +  (((this.mwtmode & MWT.DISTHOLES) != 0) ? ", dist. holes" : "")
      +  (((this.mwtmode & MWT.REVKEY)    != 0) ? ", reverse key" : "");
    System.out.println("\nsearch statistics:"
      +"\nMWT algorithm used   : " +s
      +"\nmax. recursion depth : " +this.mwt.getDepth()
      +"\nnumber of triangles  : " +this.mwt.getTriangles()
      +"\nnumber of subproblems: " +this.mwt.getSubprobs()
      +"\nnumber of splits     : " +this.mwt.getSplits()
      +"\ntotal search time    : " +(time/1000.0) +"s"
      +"\n\nsolution statistics:"
      +"\nperimeter vertices   : " +this.mwt.getVertices().length
      +"\nhole vertices        : " +this.mwt.getHoles().length
      +"\nnumber of edges      : " +n
      +"\ntriangulation weight : " +wgt);
    s = new DecimalFormat("0.000").format(wgt);
    this.stat.setText("minimum weight triangulation "
                     +"(" +n +" edge(s), weight: " +s +").");
    this.panel.repaint();       /* set status and repaint the window */
    JOptionPane.showMessageDialog(this,
      "Minimum weight triangulation has\n"
      +n +" edge(s) and weight " +s +".\n"
      +"The search took " +(time/1000.0) +" seconds\n"
      +"(statistics: " +this.mwt.getTriangles()
      +"/"             +this.mwt.getSubprobs()
      +"/"             +this.mwt.getSplits() +").",
      "Information", JOptionPane.INFORMATION_MESSAGE);
  }  /* showResult() */         /* show a success message */

  /*------------------------------------------------------------------*/
  /** Run minimum weight triangulation
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void runMWT ()
  {                             /* --- minimum weight triangulation */
    if (this.running) {         /* if running, stop computations */
      this.stat.setText("aborting computations ...");
      this.stopped = true; this.mwt.stop(); return;
    }
    if (this.pgon == null) {    /* check for a loaded pointgon */
      JOptionPane.showMessageDialog(this,
        "No pointgon has been loaded yet.",
        "Alert", JOptionPane.ERROR_MESSAGE); return;
    }
    if (!this.pgvalid) {        /* check for a valid pointgon */
      JOptionPane.showMessageDialog(this,
        "The loaded pointgon is invalid.\n"
        +"(There are duplicate vertices/holes\n"
        +"or a hole lies outside the perimeter\n"
        +"or the perimeter intersects itself.)",
        "Alert", JOptionPane.ERROR_MESSAGE); return;
    }
    if (this.runmwt == null)    /* create and show modal dialog box */
      this.runmwt = this.createRunMWT();
    this.runmwt.setVisible(true);
    if (this.mwtmode < 0)       /* check the dialog result */
      return;                   /* (< 0 means "cancelled") */
    this.setMenuState(this.running = true);
    this.stopped = false;       /* set the program state to "running" */
    this.mwt = new MWT(this.pgon);
    this.mwt.setMode(this.mwtmode);
    this.thread = new Thread(mwt); /* init. min. weight triangulation */
    this.thread.start();        /* and start computations as a thread */
    this.timer = new Timer(200, new ActionListener () {
      private int cnt = 0;      /* counter for status text update */
      public void actionPerformed (ActionEvent e) {
        if (!PGView.this.thread.isAlive()) {
          PGView.this.timer.stop();
          PGView.this.showResult();
          return;               /* stop the status update timer */
        }                       /* and show the computation result */
        if (--this.cnt <= 0) {  /* on every 5th check (once a second) */
          this.cnt = 5; PGView.this.updateStatus(); }
      } } );                    /* update the status text */
    this.timer.start();         /* start the status update timer */
  }  /* runMWT() */

  /*------------------------------------------------------------------*/
  /** Process that a mouse button was pressed.
   *  @param  me  the mouse event to process
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void mousePressed (MouseEvent me)
  {                             /* --- handle mouse clicks */
    int    m;                   /* event modifiers */
    Vertex v;                   /* buffer for a vertex */
    Edge   e;                   /* buffer for an edge */

    if (this.pgon == null) return;  /* check for a pointgon */
    this.mode = 0;              /* clear the operation mode */

    /* --- edge removal or addition --- */
    m = me.getModifiersEx();    /* get the modifiers (buttons) */
    if (me.isShiftDown()        /* if the shift key is pressed */
    ||  me.isControlDown()) {   /* or the control key is pressed */
      if ((m & InputEvent.BUTTON3_DOWN_MASK) != 0) {
        e = this.panel.findEdge(me.getX(), me.getY());
        if (e != null) { this.pgon.removeEdge(e);
          this.stat.setText("edge from vertex " +e.src.id
                           +" to vertex " +e.dst.id +" deleted.");
        } } /* remove the edge that is closest to the mouse position */
      else if ((m & InputEvent.BUTTON2_DOWN_MASK) != 0) {
        if (this.vertex != null)/* if a first vertex is known */
          this.stat.setText("vertex " +this.vertex.id +" deselected.");
        this.vertex = null; }   /* deselect already selected vertex */
      else if (this.vertex == null) { /* if there is no first vertex */
        this.vertex = this.panel.findVertex(me.getX(), me.getY());
        this.stat.setText("vertex " +this.vertex.id +" selected."); }
        /* note the vertex that is closest to the mouse position */
      else {                    /* first vertex is known */
        v = this.panel.findVertex(me.getX(), me.getY());
        if (v == this.vertex)   /* get and check second vertex */
          this.stat.setText("invalid edge.");
        else if (!this.pgon.addEdge(new Edge(this.vertex, v)))
          this.stat.setText("edge already exists.");
        else                    /* add a new inner edge */
          this.stat.setText("edge from vertex " +this.vertex.id
                           +" to vertex " +v.id +" added.");
        this.vertex = null;     /* clear the vertex buffer */
      }
      this.panel.repaint();     /* redraw the pointgon */
      return;                   /* and abort the function */
    }

    /* --- scaling or movement initialization --- */
    this.mx = me.getX();        /* note the coordinates of the point */
    this.my = me.getY();        /* at which the mouse was pressed */
    switch (m) {                /* depending on the mouse button */
      case InputEvent.BUTTON1_DOWN_MASK:
        this.mode  = 1; break;  /* panning mode */
      case InputEvent.BUTTON2_DOWN_MASK:  /* scaling (low  factor) */
      case InputEvent.BUTTON3_DOWN_MASK:  /* scaling (high factor) */
        this.mode  = (m == InputEvent.BUTTON2_DOWN_MASK) ? 2 : 3;
        this.scale = this.panel.getScale();
        this.my   -= this.scroll.getViewport().getViewPosition().y;
        break;                  /* adapt position to viewport */
    }                           /* (set the mouse operation mode) */
  }  /* mousePressed() */

  /*------------------------------------------------------------------*/
  /** Process that the mouse was dragged.
   *  @param  e  the mouse event to process
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void mouseDragged (MouseEvent e)
  {                             /* --- handle mouse movement */
    JViewport view;             /* viewport of scrollpane */
    Point     refp;             /* coordinates of upper left corner */
    Dimension size;             /* extension of viewing area */
    int       xmax, ymax, d;    /* maximum view position, movement */
    double    scl;              /* new scaling factor */

    if ((this.pgon == null)        /* check for a pointgon and */
    ||  (this.mode <= 0)) return;  /* for a valid operation mode */
    view = this.scroll.getViewport();
    refp = view.getViewPosition(); /* get reference point */
    if (this.mode == 1) {       /* if to do panning */
      refp.x += this.mx -e.getX();
      refp.y += this.my -e.getY();
      size  = this.panel.getPreferredSize();
      xmax  = size.width; ymax  = size.height;
      size  = view.getExtentSize();  /* get maximum reference point */
      xmax -= size.width; ymax -= size.height;
      if (refp.x > xmax) { this.mx -= refp.x -xmax; refp.x = xmax; }
      if (refp.x < 0)    { this.mx -= refp.x;       refp.x = 0;    }
      if (refp.y > ymax) { this.my -= refp.y -ymax; refp.y = ymax; }
      if (refp.y < 0)    { this.my -= refp.y;       refp.y = 0;    }
      view.setViewPosition(refp); }
    else {                      /* if to do scaling */
      d = (e.getY() -refp.y) -this.my;
      scl = Math.pow((this.mode > 2) ? 1.004 : 1.02, d);
      this.panel.setScale(this.scale *scl);
    }                           /* set new scaling factor */
    this.panel.revalidate();    /* adapt the enclosing scroll pane */
    this.panel.repaint();       /* and redraw the pointgon */
  }  /* mouseDragged() */

  /*------------------------------------------------------------------*/
  /** Process that a mouse button was released.
   *  @param  e  the mouse event to process
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void mouseReleased (MouseEvent e)
  { this.mode = 0; }

  /*------------------------------------------------------------------*/
  /** Process that a mouse button was pressed.
   *  @param  e  the mouse event to process
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void mouseClicked (MouseEvent e)
  { }

  /*------------------------------------------------------------------*/
  /** Process that the mouse entered the window.
   *  @param  e  the mouse event to process
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void mouseEntered (MouseEvent e)
  { }

  /*------------------------------------------------------------------*/
  /** Process that the mouse exited the window.
   *  @param  e  the mouse event to process
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void mouseExited (MouseEvent e)
  { }

  /*------------------------------------------------------------------*/
  /** Process that the mouse was moved.
   *  @param  e  the mouse event to process
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void mouseMoved (MouseEvent e)
  { }

  /*------------------------------------------------------------------*/
  /** Create the GUI of the pointgon viewer.
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void run ()
  {                             /* --- create GUI of pointgon viewer */
    JMenuBar  mbar;             /* menu bar */
    JMenu     menu;             /* to create menu titles */
    JMenuItem item;             /* to create menu items */

    this.getContentPane().setLayout(new BorderLayout());

    /* --- create and set the menu bar --- */
    mbar = new JMenuBar();
    this.getContentPane().add(mbar, BorderLayout.NORTH);
    this.mitems = new JMenuItem[4];

    menu = mbar.add(new JMenu("File"));
    menu.setMnemonic('f');
    this.mitems[1] =
    item = menu.add(new JMenuItem("Load Pointgon..."));
    item.setMnemonic('l');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        PGView.this.loadPointgon(null); } } );
    this.mitems[2] =
    item = menu.add(new JMenuItem("Reload Pointgon"));
    item.setMnemonic('r');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        PGView.this.loadPointgon(PGView.this.curr); } } );
    item = menu.add(new JMenuItem("Save Pointgon"));
    item.setMnemonic('s');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        PGView.this.savePointgon(PGView.this.curr); } } );
    item = menu.add(new JMenuItem("Save Pointgon as..."));
    item.setMnemonic('a');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        PGView.this.savePointgon(null); } } );
    item = menu.add(new JMenuItem("Save PNG Image..."));
    item.setMnemonic('i');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        PGView.this.saveImage(null); } } );
    menu.addSeparator();
    item = menu.add(new JMenuItem("Quit"));
    item.setMnemonic('q');
    if (this.isprog) {          /* if stand-alone program */
      item.addActionListener(new ActionListener() {
        public void actionPerformed (ActionEvent e) {
          System.exit(0); } } ); }     /* terminate the program */
    else {                      /* if only visualization module */
      item.addActionListener(new ActionListener() {
        public void actionPerformed (ActionEvent e) {
          if (PGView.this.about != null)
            PGView.this.about.setVisible(false);
          if (PGView.this.randpg != null)
            PGView.this.randpg.setVisible(false);
          PGView.this.setVisible(false);
        } } );                  /* only close the window */
    }                           /* and the dialog boxes */

    menu = mbar.add(new JMenu("Actions"));
    menu.setMnemonic('a');
    this.mitems[3] =
    item = menu.add(new JMenuItem("Generate Random Pointgon..."));
    item.setMnemonic('g');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        if (PGView.this.randpg == null)
          PGView.this.randpg = createRandPG();
        PGView.this.randpg.setVisible(true);
      } } );
    this.mitems[0] =
    item = menu.add(new JMenuItem("Minimum Weight Triangulation..."));
    item.setMnemonic('m');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        PGView.this.runMWT(); } } );
    item = menu.add(new JMenuItem("Compute Total Edge Weight..."));
    item.setMnemonic('w');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        if (PGView.this.pgon == null) return;
        double wgt = PGView.this.pgon.weight(true);
        PGView.this.stat.setText("total edge weight: " +wgt);
        JOptionPane.showMessageDialog(PGView.this,
          "The total edge weight of this pointgon is\n" +wgt +".",
          "Information", JOptionPane.INFORMATION_MESSAGE);
      } } );
    item = menu.add(new JMenuItem("Delete All Inner Edges"));
    item.setMnemonic('d');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        if (PGView.this.pgon == null) return;
        PGView.this.pgon.setEdges(null);
        PGView.this.panel.repaint();
        PGView.this.stat.setText("all inner edges deleted.");
      } } );
    menu.addSeparator();
    item = menu.add(new JMenuItem("Redraw"));
    item.setMnemonic('r');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        PGView.this.panel.repaint(); } } );

    menu = mbar.add(new JMenu("Help"));
    menu.setMnemonic('h');
    item = menu.add(new JMenuItem("About..."));
    item.setMnemonic('a');
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        if (PGView.this.about == null)
           PGView.this.about = PGView.this.createAbout();
        PGView.this.about.setVisible(true);
      } } );

    /* --- create and set the main panel --- */
    this.panel = new PGPanel();
    this.panel.setLayout(new BorderLayout());
    this.panel.setPreferredSize(new Dimension(400, 400));
    this.panel.addMouseListener(this);
    this.panel.addMouseMotionListener(this);
    this.scroll = new JScrollPane(this.panel);
    this.getContentPane().add(this.scroll, BorderLayout.CENTER);

    /* --- create and set a status bar --- */
    this.stat = new JTextField("");
    this.stat.setEditable(false);
    this.getContentPane().add(this.stat, BorderLayout.SOUTH);

    /* --- show the frame window --- */
    this.setTitle("PGView");
    this.setDefaultCloseOperation(this.isprog
      ? JFrame.EXIT_ON_CLOSE : JFrame.HIDE_ON_CLOSE);
    this.setLocation(50, 50);
    this.pack();
    this.stat.setText("pointgon viewer is up and running.");
  }  /* run() */

  /* Following the recommendations in the Java tutorial, the user   */
  /* interface is created in the "run" method, which is invoked     */
  /* from the event queue, in order to avoid problems with threads. */

  /*------------------------------------------------------------------*/
  /** Create a pointgon viewer.
   *  @param  isProg  whether invoked as a standalone program
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public PGView (boolean isProg)
  { this.isprog = isProg;
    try { EventQueue.invokeAndWait(this); } catch (Exception e) {} }

  /*------------------------------------------------------------------*/
  /** Create a pointgon viewer.
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public PGView ()
  { this.isprog = false;
    try { EventQueue.invokeAndWait(this); } catch (Exception e) {} }

  /*------------------------------------------------------------------*/
  /** Create a pointgon viewer.
   *  @param  title  the title of the main window
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public PGView (String title)
  { this(false); this.setTitle(title); }

  /*------------------------------------------------------------------*/
  /** Create a pointgon viewer.
   *  @param  file   the file from which to load a pointgon
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public PGView (File file)
  { this(false); this.loadPointgon(file); }

  /*------------------------------------------------------------------*/
  /** Create a pointgon viewer.
   *  @param  title  the title of the main window
   *  @param  file   the file from which to load a pointgon
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public PGView (String title, File file)
  { this(title); this.loadPointgon(file); }

  /*------------------------------------------------------------------*/
  /** Main program for command line invocation.
   *  @param  args  the command line arguments
   *  @since  2005.02.20 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static void main (String[] args)
  {                             /* --- main function */
    PGView v = new PGView(true);
    if (args.length > 0) v.loadPointgon(new File(args[0]));
    v.setVisible(true);         /* create and show a pointgon viewer */
  }  /* main() */

}  /* PGView */
