/*----------------------------------------------------------------------
  File    : Pointgon.java
  Contents: management of polygons with hole vertices
  Author  : Christian Borgelt
  History : 2005.02.18 file created
            2005.02.21 function isInside for general polygons
            2005.02.23 optional edges of partition added
            2005.02.26 functions isects and isValid added
            2005.03.02 function duplicates added, raw input format added
            2005.03.03 random pointgon generation added
            2005.03.04 functions isInside debugged
            2005.03.09 function weight added (total edge weight)
            2005.03.16 bug in isConvex fixed (angles of 180/360 degrees)
            2005.03.19 functions findVertex/Edge, add/removeEdge added
            2005.06.20 function isects corrected (segments on same line)
            2006.01.21 adapted to extended function Vertex.isectsX
            2016.04.07 StringBuffer replaced by StringBuilder
----------------------------------------------------------------------*/
package pointgon;

import java.io.InputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Random;
import util.Scanner;

/*--------------------------------------------------------------------*/
/** Class for pointgons (polygons with holes that are points).
 *  @since  2005.02.18 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public class Pointgon implements Serializable {

  private static final long serialVersionUID = 0x00010000L;
  /** block size for array enlargement */
  private static final int BLKSIZE = 32;

  /** vector of perimeter vertices */
  public  Vertex[] verts;
  /** vector of hole      vertices */
  public  Vertex[] holes;
  /** edges of partition */
  public  Edge[]   edges;
  /** current number of vertices */
  private int      vertcnt;
  /** current number of holes */
  private int      holecnt;
  /** bounding box: x-coordinate of lower left corner */
  private double   bbx;
  /** bounding box: y-coordinate of lower left corner */
  private double   bby;
  /** bounding box: width */
  private double   bbw;
  /** bounding box: height */
  private double   bbh;
  /** flag for valid bounding box */
  private boolean  valid;

  /*------------------------------------------------------------------*/
  /** Create a pointgon.
   *  @param  vertcnt  the number of perimeter vertices
   *  @param  holecnt  the number of holes
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Pointgon (int vertcnt, int holecnt)
  {                             /* --- create a pointgon */
    this.verts   = new Vertex[vertcnt];
    this.holes   = new Vertex[holecnt];
    this.vertcnt = this.holecnt = 0;
    this.edges   = null;
    this.valid   = false;
  }  /* Pointgon() */

  /*------------------------------------------------------------------*/
  /** Create a random pointgon.
   *  @param  vertcnt  the number of perimeter vertices
   *  @param  holecnt  the number of holes
   *  @param  radius   the radius parameter
   *                   (maximum perimeter vertex distance from origin)
   *  @param  rand     the random number generator
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Pointgon (int vertcnt, int holecnt, double radius, Random rand)
  {                             /* --- create a random pointgon */
    this(vertcnt, holecnt);     /* create basic structure */
    int      i, k;              /* loop variables */
    double   r;                 /* inner radius for perimeter */
    double[] phi;               /* vector of angles */
    Vertex   v;                 /* buffer for vertices */

    if      (radius < 0) radius = 0;  /* check and adapt */
    else if (radius > 1) radius = 1;  /* the radius parameter */
    phi = new double[vertcnt];  /* vector of perimeter angles */
    for (i = vertcnt; --i > 2; )/* generate random angles */
      phi[i] = rand.nextDouble() *2 *Math.PI;
    phi[0] = 0;                 /* make sure center is inside */
    phi[1] = (rand.nextDouble() +1) *0.5 *Math.PI;
    phi[2] = (rand.nextDouble() +2) *0.5 *Math.PI;
    Arrays.sort(phi,0,vertcnt); /* sort the generated angles */
    for (this.vertcnt = i = vertcnt; --i >= 0; ) {
      r = radius +(1 -radius) *rand.nextDouble();
      v = new Vertex(r *Math.cos(phi[i]), r *Math.sin(phi[i]));
      v.id = i;                 /* generate a perimeter vertex */
      this.verts[i] = v;        /* from angle and random radius */
    }                           /* and store it */
    for (this.holecnt = i = holecnt; --i >= 0; ) {
      do { v = new Vertex(2 *rand.nextDouble() -1,
                          2 *rand.nextDouble() -1);
      } while (!this.isInside(v));
      v.id = vertcnt +i;        /* generate a hole inside polygon, */
      this.holes[i] = v;        /* set the hole id and flag, */
    }                           /* and store the generated hole */
    this.valid = false;         /* invalidate bounding box */
  }  /* Pointgon() */

  /*------------------------------------------------------------------*/
  /** Add a perimeter vertex to a pointgon.
   *  @param  x     the x-coordinate of the vertex
   *  @param  y     the y-coordinate of the vertex
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void addVertex (double x, double y)
  {                             /* --- add a perimeter vertex */
    Vertex v = new Vertex(x, y);
    v.id = this.vertcnt;
    this.verts[this.vertcnt++] = v;
    this.valid = false;
  }  /* addVertex() */

  /*------------------------------------------------------------------*/
  /** Add a hole vertex to a pointgon.
   *  @param  x     the x-coordinate of the hole vertex
   *  @param  y     the y-coordinate of the hole vertex
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void addHole (double x, double y)
  {                             /* --- add a hole vertex */
    Vertex h = new Vertex(x, y);
    h.id = this.verts.length +this.holecnt;
    this.holes[this.holecnt++] = h;
  }  /* addHole() */

  /*------------------------------------------------------------------*/
  /** Set the inner edges of the pointgon (triangulation).
   *  @param  edges  the array of edges to set
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setEdges (Edge[] edges)
  {                             /* --- set inner edges */
    this.edges = edges;         /* note the new edge vector */
    if (edges == null) return;  /* if there are no edges, abort */
    for (int i = edges.length; --i >= 0; )
      edges[i].id = i;          /* set the edge identifiers */
  }  /* setEdges() */

  /*------------------------------------------------------------------*/
  /** Add an inner edge to the pointgon (triangulation).
   *  @param  edge  the edge to add
   *  @return whether the edge was added
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean addEdge (Edge edge)
  {                             /* --- add an inner edge */
    int    i, n;                /* loop variable, number of edges */
    Edge[] buf;                 /* buffer for reallocation */
    Edge   e;                   /* to traverse the edges */

    n = this.verts.length;      /* get modulo for id computations */
    if ((edge.src.id < n) && (edge.dst.id < n)
    &&  (((edge.src.id +1) % n == edge.dst.id)
    ||   ((edge.dst.id +1) % n == edge.src.id)))
      return false;             /* no edge between adjacent vertices */
    n = (this.edges != null) ? this.edges.length : 0;
    for (i = n; --i >= 0; ) {   /* get the number of edges */
      e = this.edges[i];        /* and traverse the edges */
      if (((e.src == edge.src) && (e.dst == edge.dst))
      ||  ((e.src == edge.dst) && (e.dst == edge.src)))
        return false;           /* if the edge already exists, */
    }                           /* abort with failure */
    buf = new Edge[n+1];        /* create a (new) edge vector */
    if (n > 0) System.arraycopy(this.edges, 0, buf, 0, n);
    this.edges = buf;           /* set the new edge vector */
    buf[n] = edge; edge.id = n; /* and store the new edge */
    return true;                /* return 'edge added' */
  }  /* addEdge() */

  /*------------------------------------------------------------------*/
  /** Remove an inner edge from the pointgon (triangulation).
   *  @param  edge  the edge to remove
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void removeEdge (Edge edge)
  {                             /* --- remove an inner edge */
    int    n;                   /* number of existing edges */
    Edge[] buf;                 /* buffer for reallocation */
    Edge   e;                   /* to traverse the edges */

    n = this.edges.length;      /* get the number of edges */
    if (n <= 1) {               /* if this is the last edge, */
      this.edges = null; return; } /* delete the edge vector */
    e = this.edges[n-1];        /* replace the edge to remove */
    this.edges[edge.id] = e;    /* with the last edge in the vector */
    e.id = edge.id;             /* adapt the edge idnetifier */
    buf = new Edge[n-1];        /* create a (new) edge vector */
    System.arraycopy(this.edges, 0, buf, 0, n-1);
    this.edges = buf;           /* set the new edge vector */
  }  /* removeEdge() */

  /*------------------------------------------------------------------*/
  /** Resize a pointgon (change number of vertices).
   *  @param  vertcnt  the new number of perimeter vertices
   *  @param  holecnt  the new number of holes
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void resize (int vertcnt, int holecnt)
  {                             /* --- resize a pointgon */
    int      i;                 /* loop variable */
    Vertex[] buf;               /* buffer for reallocation */

    if (holecnt != this.holes.length) {
      buf = new Vertex[holecnt];
      if (holecnt < this.holecnt) this.holecnt = holecnt;
      System.arraycopy(this.holes, 0, buf, 0, this.holecnt);
      this.holes = buf;         /* resize and copy the holes vector */
      if (vertcnt != this.verts.length) {
        for (i = this.holecnt; --i >= 0; )
          buf[i].id = vertcnt +i;
      }                         /* set new hole identifiers */
    }
    if (vertcnt != this.verts.length) {
      buf = new Vertex[vertcnt];
      if (vertcnt < this.vertcnt) this.vertcnt = vertcnt;
      System.arraycopy(this.verts, 0, buf, 0, this.vertcnt);
      this.verts = buf;         /* resize the vertex vector and */
    }                           /* copy the existing vertices */
    this.valid = false;         /* invalidate bounding box */
  }  /* resize() */

  /*------------------------------------------------------------------*/
  /** Scale and shift a pointgon.
   *  @param  scale    the scaling parameter
   *  @param  xoff     the x-offset by which to shift
   *  @param  yoff     the y-offset by which to shift
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void transform (double scale, double xoff, double yoff)
  {                             /* --- scale a pointgon */
    int    i;                   /* loop variable */
    Vertex v;                   /* to traverse the vertices */

    for (i = this.verts.length; --i >= 0; ) {
      v = this.verts[i];        /* traverse the perimeter vertices */
      v.x = v.x *scale +xoff;   /* and transform their coordinates */
      v.y = v.y *scale +yoff;
    }
    for (i = this.holes.length; --i >= 0; ) {
      v = this.holes[i];        /* traverse the hole vertices */
      v.x = v.x *scale +xoff;   /* and transform their coordinates */
      v.y = v.y *scale +yoff;
    }
    this.valid = false;         /* bounding box is no longer valid */
  }  /* transform() */

  /*------------------------------------------------------------------*/
  /** Compute a bounding box of a pointgon.
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void bbox ()
  {                             /* --- compute bounding box */
    int    i;                   /* loop variable */
    double x, y;                /* coordinates of a vertex */
    double xmax, ymax;          /* maximal x- and y-coordinates */

    this.bbx = Double.MAX_VALUE; xmax = -Double.MAX_VALUE;
    this.bby = Double.MAX_VALUE; ymax = -Double.MAX_VALUE;
    for (i = this.verts.length; --i >= 0; ) {
      x = this.verts[i].x;      /* traverse the vertices */
      y = this.verts[i].y;      /* of the perimeter */
      if (x < this.bbx) this.bbx = x;
      if (x > xmax)     xmax     = x;
      if (y < this.bby) this.bby = y;
      if (y > ymax)     ymax     = y;
    }                           /* find minimum and maximum coords. */
    this.bbw = xmax -this.bbx;  /* compute the width and height */
    this.bbh = ymax -this.bby;  /* of the bounding box */
    this.valid = true;          /* the bounding box is now valid */
  }  /* bbox() */

  /*------------------------------------------------------------------*/
  /** Get the x-coordinate of lower left corner of the bounding box.
   *  @return the x-coordinate of lower left corner of the bounding box
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getX ()
  { if (!this.valid) this.bbox(); return this.bbx; }

  /*------------------------------------------------------------------*/
  /** Get the y-coordinate of lower left corner of the bounding box.
   *  @return the y-coordinate of lower left corner of the bounding box
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getY ()
  { if (!this.valid) this.bbox(); return this.bby; }

  /*------------------------------------------------------------------*/
  /** Get the width the bounding box.
   *  @return the width of the bounding box
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getWidth ()
  { if (!this.valid) this.bbox(); return this.bbw; }

  /*------------------------------------------------------------------*/
  /** Get the height of the bounding box.
   *  @return the height of the bounding box
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getHeight ()
  { if (!this.valid) this.bbox(); return this.bbh; }

  /*------------------------------------------------------------------*/
  /** Reverse (the order of) the perimeter vertices.
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void reverse ()
  {                             /* --- reverse perimeter */
    int    l, r;                /* loop variables */
    Vertex a, b;                /* buffers for exchange */

    for (l = 0, r = this.verts.length-1; l < r; l++, r--) {
      a = this.verts[l]; a.id = r;
      b = this.verts[r]; b.id = l;
      this.verts[l] = b;        /* mirror the list of */
      this.verts[r] = a;        /* perimeter vertices */
    }                           /* and set new vertex ids */
  }  /* reverse() */

  /*------------------------------------------------------------------*/
  /** Check whether the pointgon is convex.
   *  @return whether the pointgon is convex
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isConvex ()
  {                             /* --- check for convexity */
    int     i;                  /* loop variable */
    Vertex  p, q;               /* to traverse the perimeter vertices */
    double  u, v, x, y, ox, oy; /* coordinates of direction vectors */
    boolean c = false;          /* flag for counterclockwise order */

    if (this.verts.length < 4)  /* all pointgons with less than four */
      return true;              /* perimeter vertices are convex */
    p  = this.verts[i = this.verts.length-1];
    q  = this.verts[0];         /* get the wrap around edge and */
    ox = q.x -p.x;              /* compute the coordinates of the */
    oy = q.y -p.y;              /* direction vector of this edge */
    while (--i >= 0) {          /* traverse the remaining edges */
      q = p; p = this.verts[i]; /* compute direction vector of the */
      x = q.x -p.x; v = oy *x;  /* next edge and the scalar product */
      y = q.y -p.y; u = ox *y;  /* with the rotated previous one */
      if (u != v) {             /* if the angle between consecutive */
        c = (u >= v); break; }  /* edges is != 180 / 360 degrees, */
      if ((x != 0) || (y != 0)){/* init. the order flag and abort, */
        ox = x; oy = y; }       /* otherwise go to the next edge */
    }
    p  = this.verts[i = this.verts.length-1];
    q  = this.verts[0];         /* get the wrap around edge and */
    ox = q.x -p.x;              /* compute the coordinates of the */
    oy = q.y -p.y;              /* direction vector of this edge */
    while (--i >= 0) {          /* traverse the remaining edges */
      q = p; p = this.verts[i]; /* compute direction vector of the */
      x = q.x -p.x; v = oy *x;  /* next edge and the scalar product */
      y = q.y -p.y; u = ox *y;  /* with the rotated previous one */
      if (((u != v)             /* if the angle between two */
      &&  ((u >= v) != c))      /* consecutive perimeter edges */
      ||  ((u == v)             /* is greater than 180 degrees */
      &&  ((ox *x < 0) || (oy *y < 0))))
        return false;           /* the polygon is not convex */
      if ((x != 0) || (y != 0)) {
        ox = x; oy = y; }       /* only copy the direction vector */
    }                           /* if it is not the null vector */
    return true;                /* return 'pointgon is convex' */
  }  /* isConvex() */

  /*------------------------------------------------------------------*/
  /** Check whether a vertex is inside the pointgon.
   *  @param  v   the vertex to check
   *  @return whether the vertex is inside the pointgon
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isInside (Vertex v)
  {                             /* --- check whether vertex is inside */
    int    i, r, n = 0;         /* loop variable, buffer, counter */
    Vertex p, q;                /* perimeter vertices */

    if (this.verts.length < 3)  /* all pointgons with less than */
      return false;             /* three vertices have no interior */
    p = this.verts[0];          /* get vertex for wrap around */
    for (i = this.verts.length; --i >= 0; ) {
      q = p; p = this.verts[i]; /* traverse the perimeter edges (p,q) */
      r = v.horzLineCuts(p, q); /* check whether horz. line cuts */
      if (r < -1) continue;     /* if edge is out of range, skip it */
      if (r == 0) return false; /* if vertex is on the edge, abort */
      if (r >  0) n++;          /* if horz. line cuts to the right */
    }                           /* of the vertex, count the edge */
    /* The idea of the above check is to shoot a straight line from  */
    /* the given point to the right and to count how many times this */
    /* line cuts the perimeter. If the number is odd, the point lies */
    /* inside the perimeter, if it is even, the point lies outside.  */
    return (n & 1) != 0;        /* check the number of intersections */
  }  /* isInside() */           /* (odd: inside, even: outside) */

  /*------------------------------------------------------------------*/
  /** Check whether all holes are inside the pointgon.
   *  @return whether all holes are inside the pointgon
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean allHolesInside ()
  {                             /* --- check whether holes are inside */
    for (int i = this.holes.length; --i >= 0; )
      if (!this.isInside(this.holes[i]))
        return false;           /* if a hole is outside, abort */
    return true;                /* return 'all holes are inside' */
  }  /* allHolesInside() */

  /*------------------------------------------------------------------*/
  /** Check for duplicate holes.
   *  @return whether there are duplicate holes
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean duplicates ()
  {                             /* --- check for duplicate holes */
    int      i;                 /* loop variable */
    Vertex[] vec;               /* buffer for holes */
    Vertex   u, v;              /* vertices */

    v = this.verts[0];          /* get first vertex for wrap around */
    for (i = this.verts.length; --i >= 0; ) {
      u = v; v = this.verts[i]; /* traverse the perimeter vertices */
      if ((u.x == v.x) && (u.y == v.y))
        return true;            /* if consecutive perimeter vertices */
    }                           /* are identical, abort with duplic. */
    if (this.holes.length <  2) /* if there are less than two holes, */
      return false;             /* there cannot be any duplicates */
    if (this.holes.length == 2) /* special check for two holes */
      return (this.holes[0].compareTo(this.holes[1]) == 0);
    vec = new Vertex[i = this.holes.length];
    while (--i >= 0)            /* create a buffer for the holes */
      vec[i] = this.holes[i];   /* and copy the holes into it, */
    Arrays.sort(vec, 0, vec.length);   /* then sort the vector */
    for (v = vec[i = vec.length-1]; --i >= 0; ) {
      u = v; v = vec[i];        /* traverse the sorted holes */
      if ((u.x == v.x) && (u.y == v.y))
        return true;            /* if consecutive hole vertices are */
    }                           /* identical, abort with duplicates */
    return false;               /* return 'no duplicates' */
  }  /* duplicates() */

  /*------------------------------------------------------------------*/
  /** Check for perimieter intersections.
   *  @return whether there are perimeter intersections
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isects ()
  {                             /* --- check perimeter intersections */
    int    i, k, n = 1;         /* loop variables, buffer */
    Vertex s1, s2, d1, d2;      /* to traverse the perimeter vertices */

    s1 = this.verts[0];         /* get vertex for wrap around */
    for (i = this.verts.length; --i > 1; n = 0) {
      d1 = s1;                  /* traverse the vertices */
      s1 = this.verts[i];       /* on the pointgon's perimeter */
      for (s2 = this.verts[k = i-1]; --k >= n; ) {
        d2 = s2;                /* traverse the vertices */
        s2 = this.verts[k];     /* on the pointgon's perimeter */
        if (s1.isectsX(d1, s2, d2))
          return true;          /* if the perimeter intersects */
      }                         /* with itself, abort the function */
    }
    return false;               /* return 'no intersection' */
  }  /* isects() */

  /*------------------------------------------------------------------*/
  /** Check for a valid state of the pointgon.
   *  <p>A pointgon is valid if there are no duplicate holes,
   *  all holes are inside the pointgon and there are no perimeter
   *  intersections.
   *  @return whether there are perimeter intersections
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean isValid ()
  {                             /* --- check for a valid state */
    if (this.duplicates()) {    /* check for duplicate vertices */
      System.out.println("duplicate vertices/holes");    return false; }
    if (!this.allHolesInside()){/* check whether holes are inside */
      System.out.println("not all holes inside");        return false; }
    if (this.isects()) {        /* check for perimeter intersections */
      System.out.println("perimeter intersects itself"); return false; }
    return true;                /* return 'pointgon is valid' */
  }  /* isValid() */

  /*------------------------------------------------------------------*/
  /** Compute the total weight of the edges.
   *  @param  withPerimeter  whether to include the perimeter edges
   *  @return the total weight of the edges
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double weight (boolean withPerimeter)
  {                             /* --- compute weight of edges */
    int    i;                   /* loop variable */
    Vertex p, q;                /* to traverse the perimeter */
    double wgt = 0;             /* total weight of edges */

    if (this.edges != null)     /* sum the edge lengths */
      for (i = this.edges.length; --i >= 0; )
        wgt += this.edges[i].wgt;
    if (withPerimeter) {        /* if to add perimeter edges */
      p = this.verts[0];        /* get first vertex for warp around */
      for (i = this.verts.length; --i >= 0; ) {
        q = p; p = this.verts[i];
        wgt += p.distance(q);   /* sum the lengths */
      }                         /* of the perimeter edges */
    }
    return wgt;                 /* return the total edge weight */
  }  /* weight() */

  /*------------------------------------------------------------------*/
  /** Find the closest vertex to a given point.
   *  @param  x    the x-coordinate of the point
   *  @param  y    the y-coordinate of the point
   *  @return the closest vertex to the point
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Vertex findVertex (double x, double y)
  {                             /* --- find closest vertex */
    int    i;                   /* loop variable */
    double d, min;              /* (minimal) distance */
    Vertex v = null;            /* closest vertex */

    min = Double.MAX_VALUE;     /* traverse the perimeter vertices */
    for (i = this.verts.length; --i >= 0; ) {
      d = this.verts[i].distance(x, y);
      if (d <= min) { min = d; v = this.verts[i]; }
    }                           /* find the closest perimeter vertex */
    if (this.holes == null)     /* if there are no holes, */
      return v;                 /* return the perimeter vertex */
    for (i = this.holes.length; --i >= 0; ) {
      d = this.holes[i].distance(x, y);
      if (d <= min) { min = d; v = this.holes[i]; }
    }                           /* check for a hole that is closer */
    return v;                   /* return the closest vertex */
  }  /* findVertex() */

  /*------------------------------------------------------------------*/
  /** Find the closest edge to a given point.
   *  @param  x    the x-coordinate of the point
   *  @param  y    the y-coordinate of the point
   *  @return the closest edge to the point
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Edge findEdge (double x, double y)
  {                             /* --- find closest edge */
    int    i;                   /* loop variable */
    double d, min;              /* (minimal) distance */
    Edge   e = null;            /* closest edge */

    if (this.edges == null)     /* if there are no edges, */
      return null;              /* abort the function */
    min = Double.MAX_VALUE;     /* traverse the perimeter vertices */
    for (i = this.edges.length; --i >= 0; ) {
      d = this.edges[i].distance(x, y);
      if (d <= min) { min = d; e = this.edges[i]; }
    }                           /* find the closest edge */
    return e;                   /* and return it */
  }  /* findEdge() */

  /*------------------------------------------------------------------*/
  /** Create a string description of the pointgon.
   *  @return the created string description
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public String toString ()
  {                             /* --- create a string description */
    int           i;            /* loop variable */
    StringBuilder s;            /* created string description */

    s = new StringBuilder("pointgon = {\n");
    s.append("  vertices = {\n    " +this.verts[0]);
    for (i = 1; i < this.verts.length; i++)
      s.append(",\n    " +this.verts[i]);
    s.append("\n  };\n");       /* list the perimeter vertices */
    if (this.holes.length > 0){ /* if there are hole vertices */
      s.append("  holes    = {\n    " +this.holes[0]);
      for (i = 1; i < this.holes.length; i++)
        s.append(",\n    " +this.holes[i]);
      s.append("\n  };\n");     /* list the hole vertices */
    }
    if ((this.edges != null)    /* if there are edges */
    &&  (this.edges.length > 0)) {
      s.append("  edges    = {\n    " +this.edges[0]);
      for (i = 1; i < this.edges.length; i++)
        s.append(",\n    " +this.edges[i]);
      s.append("\n  };\n");     /* list the edges */
    }
    s.append("};\n");           /* terminate the description */
    return s.toString();        /* return the created string */
  }  /* toString() */

  /*------------------------------------------------------------------*/
  /** Read a list of vertices.
   *  @param  s   the scanner to read from
   *  @return the read list of vertices
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private static Vertex[] readRaw (Scanner s) throws IOException
  {                             /* --- read a list of vertices */
    double   x, y;              /* coordinates of a vertex */
    int      i = 0, n;          /* number of vertices, buffer */
    Vertex[] vec, buf;          /* vertex vector, buffer for realloc. */

    vec = new Vertex[n = BLKSIZE];  /* create an initial vector */
    while (s.nextToken() == Scanner.T_NUM) {
      x = Double.parseDouble(s.value);
      s.getNumber();            /* get the y-coordinate */
      y = Double.parseDouble(s.value);
      s.getChar('\n');          /* check for a newline */
      if (i >= n) {             /* if the vertex vector is full */
        n  += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
        buf = new Vertex[n];    /* create a new vertex vector */
        System.arraycopy(vec, 0, buf, 0, i);
        vec = buf;              /* copy the vertices into it */
      }                         /* and set the new vector */
      vec[i++] = new Vertex(x, y);    /* store the vertex */
    }
    if ((s.ttype != '\n')       /* check for an empty line */
    &&  (s.ttype != Scanner.T_EOF))      /* or end of file */
      throw new IOException("number or '\\n' expected instead of \""
                            +s.value +"\"" +s.lno());
    if (i == 0)                 /* if nothing was read, */
      return new Vertex[0];     /* return an empty vector */
    if (i < vec.length) {       /* if the vertex vector is not full */
      buf = new Vertex[i];      /* create a new vertex vector */
      System.arraycopy(vec, 0, buf, 0, i);
      vec = buf;                /* copy the vertices into it */
    }                           /* and set the new vector */
    return vec;                 /* return the vertex vector */
  }  /* readRaw() */

  /*------------------------------------------------------------------*/
  /** Read a list of vertices.
   *  @param  s   the scanner to read from
   *  @return the read list of vertices
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private static Vertex[] readVertices (Scanner s) throws IOException
  {                             /* --- read a list of vertices */
    double   x, y;              /* coordinates of a vertex */
    int      i = 0, n;          /* number of vertices, buffer */
    Vertex[] vec, buf;          /* vertex vector, buffer for realloc. */

    s.getChar('{');             /* check for a '{' */
    if (s.nextToken() == '}')   /* if the list is empty, */
      return new Vertex[0];     /* return an empty vector */
    s.pushBack();               /* push back already read token */
    vec = new Vertex[n = BLKSIZE];  /* create an initial vector */
    do {                        /* vertex read loop */
      s.getChar('(');           /* check for a '(' */
      s.getNumber();            /* get the x-coordinate */
      x = Double.parseDouble(s.value);
      s.getChar(',');           /* check for a ',' */
      s.getNumber();            /* get the y-coordinate */
      y = Double.parseDouble(s.value);
      s.getChar(')');           /* check for a ')' */
      if (i >= n) {             /* if the vertex vector is full */
        n  += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
        buf = new Vertex[n];    /* create a new vertex vector */
        System.arraycopy(vec, 0, buf, 0, i);
        vec = buf;              /* copy the vertices into it */
      }                         /* and set the new vector */
      vec[i++] = new Vertex(x, y);    /* store the vertex */
    } while (s.nextToken() == ',');
    s.pushBack();               /* push back last token */
    s.getChar('}');             /* check for a '}' */
    if (i < vec.length) {       /* if the vertex vector is not full */
      buf = new Vertex[i];      /* create a new vertex vector */
      System.arraycopy(vec, 0, buf, 0, i);
      vec = buf;                /* copy the vertices into it */
    }                           /* and set the new vector */
    return vec;                 /* return the vertex vector */
  }  /* readVertices() */

  /*------------------------------------------------------------------*/
  /** Read a list of edges of a pointgon.
   *  @param  s     the scanner to read from
   *  @param  verts the perimeter vertices of the pointgon
   *  @param  holes the hole      vertices of the pointgon
   *  @return the read list of edges
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private static Edge[] readEdges (Scanner s, Vertex[] verts,
                                              Vertex[] holes)
    throws IOException
  {                             /* --- read a list of edges */
    int    i = 0, k, n;         /* number of vertices, buffers */
    int    cnt;                 /* total number of vertices */
    Vertex src, dst;            /* source and destination vertex */
    Edge[] vec, buf;            /* vertex vector, buffer for realloc. */

    s.getChar('{');             /* check for a '{' */
    if (s.nextToken() == '}')   /* if the list is empty, */
      return null;              /* return a null vector */
    s.pushBack();               /* push back already read token */
    vec = new Edge[n = BLKSIZE];/* create an initial vector */
    cnt = verts.length +holes.length;
    do {                        /* edge read loop */
      s.getChar('(');           /* check for a '(' */
      s.getNumber();            /* get the source index */
      k = Integer.parseInt(s.value);
      if ((k < 0) || (k >= cnt))
        throw new IOException("invalid vertex index " +k +s.lno());
      src = (k < verts.length) ? verts[k] : holes[k-verts.length];
      s.getChar(',');           /* check for a ',' */
      s.getNumber();            /* get the destination index */
      k = Integer.parseInt(s.value);
      if ((k < 0) || (k >= cnt))
        throw new IOException("invalid vertex index " +k +s.lno());
      dst = (k < verts.length) ? verts[k] : holes[k-verts.length];
      s.getChar(')');           /* check for a ')' */
      if (i >= n) {             /* if the edge vector is full */
        n  += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
        buf = new Edge[n];      /* create a new edge vector */
        System.arraycopy(vec, 0, buf, 0, i);
        vec = buf;              /* copy the edges into it */
      }                         /* and set the new vector */
      vec[i++] = new Edge(src, dst);    /* store the edge */
    } while (s.nextToken() == ',');
    s.pushBack();               /* push back last token */
    s.getChar('}');             /* check for a '}' */
    if (i < vec.length) {       /* if the edge vector is not full */
      buf = new Edge[i];        /* create a new edge vector */
      System.arraycopy(vec, 0, buf, 0, i);
      vec = buf;                /* copy the edges into it */
    }                           /* and set the new vector */
    return vec;                 /* return the edge vector */
  }  /* readEdges() */

  /*------------------------------------------------------------------*/
  /** Read a pointgon from a scanner.
   *  @param  s     the scanner to read from
   *  @throws IOException if a read error occcurred
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Pointgon (Scanner s) throws IOException
  {                             /* --- read a pointgon */
    int     i, n;               /* loop variable */
    boolean raw;                /* flag for raw input format */

    raw = (s.nextToken() == Scanner.T_NUM);
    s.pushBack();               /* look ahead to the first token */

    /* --- read perimeter vertices --- */
    if (raw) {                  /* if the file is in raw format */
      s.nlToken(true);          /* make newline a token */
      this.verts = readRaw(s); }
    else {                      /* if in sophisticated format */
      s.getID("pointgon");      /* check for keyword 'pointgon' */
      s.getChar('=');           /* check for a '=' */
      s.getChar('{');           /* check for a '{' */
      s.getID("vertices");      /* check for keyword 'vertices' */
      s.getChar('=');           /* check for a '=' */
      this.verts = readVertices(s);
      s.getChar(';');           /* read the vertex list */
    }                           /* and check for a ';' */
    if (this.verts.length < 3)  /* if there are too few vertices */
      throw new IOException("too few perimeter vertices" +s.lno());
    this.vertcnt = n = i = this.verts.length;
    while (--i >= 0)            /* number the perimeter vertices */
      this.verts[i].id = i;     /* from 0 to n-1 */

    /* --- read hole vertices --- */
    if (raw)                    /* if the file is in raw format */
      this.holes = readRaw(s);
    else {                      /* if in sophisticated format */
      if ((s.nextToken() != Scanner.T_ID)
      ||  !s.value.equals("holes")) {
        this.holes = new Vertex[0];  /* if there is no list of holes, */
        s.pushBack(); }              /* set an empty vector, */
      else {                    /* if there is a  list of holes */
        s.getChar('=');         /* check for a '=' */
        this.holes = readVertices(s);
        s.getChar(';');         /* read the vertex list */
      }                         /* and check for a ';' */
    }
    this.holecnt = i = this.holes.length;
    while (--i >= 0)            /* number the hole vertices */
      this.holes[i].id = n+i;   /* from n to n+k-1 */

    /* --- read edges --- */
    if (!raw) {                 /* if in sophisticated format */
      if ((s.nextToken() != Scanner.T_ID)
      ||  !s.value.equals("edges")) {
        this.edges = null;      /* if there is no list of edges, */
        s.pushBack(); }         /* set a null vector, */
      else {                    /* if there is a  list of edges */
        s.getChar('=');         /* check for a '=' */
        this.edges = readEdges(s, this.verts, this.holes);
        s.getChar(';');         /* read the edge list and */
      }                         /* check for a ';' */
      s.getChar('}');           /* check for a '}' */
      s.getChar(';');           /* check for a ';' */
    }
    this.valid = false;         /* invalidate the bounding box */
  }  /* Pointgon() */

  /*------------------------------------------------------------------*/
  /** Read a pointgon from a string description.
   *  @param  desc   the string description to read from
   *  @throws IOException if a read error occcurred
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Pointgon (String desc)    throws IOException
  { this(new Scanner(desc)); }

  /*------------------------------------------------------------------*/
  /** Read a pointgon from an input stream.
   *  @param  in    the input stream to read from
   *  @throws IOException if a read error occcurred
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Pointgon (InputStream in) throws IOException
  { this(new Scanner(in)); }

  /*------------------------------------------------------------------*/
  /** Main program for command line invocation.
   *  @param  args  the command line arguments
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static void main (String[] args)
  {                             /* --- main function for testing */
    int      vertcnt;           /* number of perimeter vertices */
    int      holecnt = 0;       /* number of hole vertices */
    double   radius  = 0.2;     /* inner radius (as a fraction) */
    double   scale   = 12.0;    /* scaling factor */
    long     seed;              /* seed value for random numbers */
    Pointgon pgon;              /* created random pointgon */

    if (args.length <= 0) {     /* if no file name is given */
      System.err.println("usage: Pointgon <vertcnt> [<holecnt>] "
        +"[<radius>] [<scale>] [<seed>]"); return; }
    seed    = System.currentTimeMillis();
    vertcnt = Integer.parseInt(args[0]);
    if (args.length > 1) holecnt = Integer.parseInt(args[1]);
    if (args.length > 2) radius  = Integer.parseInt(args[2]) /100.0;
    if (args.length > 3) scale   = Double.parseDouble(args[3]);
    if (args.length > 4) seed    = Integer.parseInt(args[4]);
    pgon = new Pointgon(vertcnt, holecnt, radius, new Random(seed));
    pgon.transform(scale,0,0);  /* create a pointgon and scale it */
    System.out.print(pgon);     /* print the created pointgon */
  }  /* main() */

}  /* class Pointgon */
