/*----------------------------------------------------------------------
  File    : Splitter.java
  Contents: split a (sub-)pointgon into (sub-)pointgons
  Author  : Christian Borgelt
  History : 18.02.2005 file created (as part of file MWT.java)
            28.12.2005 class made abstract, old class is PathSplitter
            04.01.2006 parameter mwt of init method moved to constructor
            21.01.2006 function show added for debugging purposes
            27.01.2006 triangle moved here from subclasses
            06.02.2006 some general functions moved here
----------------------------------------------------------------------*/
package pointgon;

import java.io.Serializable;

/*--------------------------------------------------------------------*/
/** Class for splitting a (sub-)pointgon into (sub-)pointgons.
 *  @since  2006.01.27 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public abstract class Splitter implements Serializable {

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  private static final long serialVersionUID = 0x00010000;

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** all (input) perimeter vertices */
  protected Vertex[]   verts;
  /** order of perimeter vertices */
  protected int        dir;
  /** table of edge weights/validity */
  protected double[][] wgts;
  /** flag table for valid triangles */
  protected char[][][] trtab;
  /** whether to distribute holes */
  protected boolean    dist;
  /** key of subproblem */
  protected Vertex[]   key;
  /** key of right subproblem */
  protected Vertex[]   rgtkey;
  /** key of left  subproblem */
  protected Vertex[]   lftkey;
  /** size of subproblem */
  protected int        cnt;
  /** size of right subproblem */
  protected int        rgtcnt;
  /** size of left  subproblem */
  protected int        lftcnt;
  /** holes in subproblem */
  protected Vertex[]   holes;
  /** holes in right subproblem */
  protected Vertex[]   rgtholes;
  /** holes in left  subproblem */
  protected Vertex[]   lftholes;
  /** number of holes in subproblem */
  protected int        holecnt;
  /** number of holes in right subproblem */
  protected int        rgtholecnt;
  /** number of holes in left  subproblem */
  protected int        lftholecnt;
  /** triangle cut out of the subproblem */
  protected Vertex[]   triangle;
  /** state of the splitter */
  protected int        state;
  /** best split of the subproblem */
  protected transient Object best;

  /*------------------------------------------------------------------*/
  /** Create a subproblem splitter.
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Splitter ()
  { }

  /*------------------------------------------------------------------*/
  /** Check a triangle.
   *  @param  a  first  vertex of triangle
   *  @param  b  second vertex of triangle
   *  @param  c  third  vertex of triangle
   *  @return whether the triangle is valid
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected boolean valid (Vertex a, Vertex b, Vertex c)
  {                             /* --- check a triangle */
    int    i;                   /* loop variable */
    Vertex h;                   /* to traverse the hole vertices */

    if (this.trtab != null)     /* if there is precomputed info., */
      return (c.isLeftOf(a, b)  /* check the triangle with it */
          &&  (this.trtab[a.id][b.id][c.id] != 0));
    if ((this.wgts[a.id][c.id] < 0)
    ||  (this.wgts[b.id][c.id] < 0)
    ||  !c.isLeftOf(a, b))      /* check for a proper triangle */
      return false;             /* with valid edges to the tip */
    for (i = this.holecnt; --i >= 0; ) {
      h = this.holes[i];        /* traverse the hole vertices */
      if ((h != c) && h.isInside(a, b, c))
        return false;           /* if the triangle contains a hole, */
    }                           /* the triangle is invalid */
    /* Note that it suffices to check whether the vertex c is equal */
    /* to a hole, since how this function is called already ensures */
    /* vertices a and b cannot be in the list of holes tested here. */
    return true;                /* otherwise the triangle is valid */
  }  /* valid() */

  /*------------------------------------------------------------------*/
  /** Check intersection with key.
   *  @param  src  the source      vertex
   *  @param  dst  the destination vertex
   *  @return whether the edges between the vertices intersects path
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected boolean isects (Vertex src, Vertex dst)
  {                             /* --- check intersection with key */
    int    i;                   /* loop variable */
    Vertex p, q;                /* to traverse the path vertices */

    q = this.key[0];            /* start at the subproblem anchor */
    for (i = this.cnt; --i >= 0; q = p) {
      p = this.key[i];          /* traverse the edges of the paths */
      if ((p != null) && (p != src) && (p != dst)
      &&  (q != null) && (q != src) && (q != dst)
      &&  p.isectsX(q,src,dst)) /* if the given edge intersects */
        return true;            /* an edge of the subproblem key, */
    }                           /* return 'edge intersects path' */
    return false;               /* otherwise return 'no intersection' */
  }  /* isects() */

  /*------------------------------------------------------------------*/
  /** Initialize a subproblem splitter.
   *  @param  key      the key  of the subproblem
   *  @param  cnt      the size of the subproblem
   *  @param  holes    the holes of the subproblem
   *  @param  holecnt  the number of holes of the subproblem
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected abstract void init (Vertex[] key,   int cnt,
                                Vertex[] holes, int holecnt);

  /*------------------------------------------------------------------*/
  /** Initialize a subproblem splitter.
   *  @param  key      the key  of the subproblem
   *  @param  cnt      the size of the subproblem
   *  @param  spobj    the splitter object
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected abstract void init (Vertex[] key, int cnt, Object spobj);

  /*------------------------------------------------------------------*/
  /** Check for an empty triangle.
   *  @return whether the traingle is empty
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected double empty ()
  {                             /* --- check for an empty triangle */
    int    n;                   /* number of perimeter vertices */
    Vertex v, p, q;             /* vertices for triangle check */

    if ((this.holecnt > 0)      /* if there are holes or there */
    ||  (this.cnt     > 3))     /* are more than three vertices, */
      return -1.0;              /* this is not an empty triangle */
    v = this.key[0];            /* get the anchor of the subproblem */
    p = this.key[1];            /* and the counterclockwise and */
    q = this.key[2];            /* clockwise adjacent vertices */
    if ((p == null) || (q == null)) {
      n = this.verts.length;    /* get modulo for id computations */
      if (p == null) p = this.verts[(v.id +n +this.dir) % n];
      if (q == null) q = this.verts[(v.id +n -this.dir) % n];
      if ((p.id +n +this.dir) % n != q.id)
        return -1.0;            /* check for adjacent vertices */
    }                           /* (otherwise this is not a triangle) */
    return this.wgts[v.id][p.id]/* if this is actually a triangle, */
         + this.wgts[v.id][q.id]/* return its weight */
         + this.wgts[p.id][q.id];
  }  /* empty() */

  /*------------------------------------------------------------------*/
  /** Create next subproblem split.
   *  @return whether a new split was created
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected abstract boolean next ();

  /*------------------------------------------------------------------*/
  /** Split a subproblem with the current split.
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected abstract void split ();

  /*------------------------------------------------------------------*/
  /** Get the current triangle.
   *  @return the current triangle
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected abstract Vertex[] triangle ();

  /*------------------------------------------------------------------*/
  /** Compute the weight of the current split.
   *  @return the weight of a split
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected abstract double weight ();

  /*------------------------------------------------------------------*/
  /** Get the edges of the current path.
   *  @return the edges of the current path as an array
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected abstract Edge[] edges ();

  /*------------------------------------------------------------------*/
  /** Store the current path as best.
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected abstract void store ();

  /*------------------------------------------------------------------*/
  /** Retrieve the best path.
   *  @return the best path
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected abstract Object retrieve ();

  /*------------------------------------------------------------------*/
  /** Distribute the hole vertices.
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected abstract void distribute ();

  /*------------------------------------------------------------------*/
  /** Show the subproblem to process.
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void show () { this.show(0); }

  /*------------------------------------------------------------------*/
  /** Show the subproblem to process.
   *  @param  depth  the recursion depth
   *  @since  2006.01.27 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  protected void show (int depth)
  {                             /* --- show subproblem to process */
    for (int i = depth; --i >= 0; )
      System.out.print("  ");   /* indent the output line */
    System.out.print("key  :"); /* print the subproblem key */
    for (int i = 0; i < this.cnt; i++) {
      System.out.print(" ");    /* print a separator */
      if (this.key[i] != null) System.out.print(this.key[i].id +":");
      System.out.print(this.key[i]);
    }                           /* traverse and print key elements */
    System.out.println();       /* terminate the output line */
    for (int i = depth; --i >= 0; )
      System.out.print("  ");   /* indent the output line */
    System.out.print("holes:"); /* print the list of holes */
    for (int i = 0; i < this.holecnt; i++)
      System.out.print(" " +this.holes[i].id +":" +this.holes[i]);
    System.out.println();       /* traverse and print the holes */
  }  /* show() */               /* terminate the output line */

}  /* class Splitter */
