/*----------------------------------------------------------------------
  File    : Trie.java
  Contents: Trie structure to store partition subproblem solutions
  Author  : Christian Borgelt
  History : 2005.02.18 file created
            2005.03.04 adapted to changed key format
            2005.03.11 separate look ups for getWeight and getPath
            2005.05.11 generalized to storing paths and vertices
            2005.12.28 specialized functions (for path/tip) removed
            2006.01.04 instance variables of trie made private
            2006.02.19 direction of key use made an option
----------------------------------------------------------------------*/
package pointgon;

import java.io.Serializable;

/*--------------------------------------------------------------------*/
/** Class for a trie node.
 *  @author Christian Borgelt
 *  @since  2005.02.18 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
class TrieNode implements Serializable {

  private static final long serialVersionUID = 0x00010000L;

  /** weight of partition */
  protected double     weight;
  /** splitting path/vertex */
  protected transient Object object;
  /** vector of child nodes */
  protected TrieNode[] children;

  /*------------------------------------------------------------------*/
  /** Create a trie node.
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public TrieNode ()
  { this.weight = -1.0; this.object = null; this.children = null; }

}  /* class TrieNode */


/*--------------------------------------------------------------------*/
/** Class for a trie to store partition subproblem solutions.
 *  @author Christian Borgelt
 *  @since  2005.02.18 (Christian Borgelt) */
/*--------------------------------------------------------------------*/
public class Trie implements Serializable {

  private static final long serialVersionUID = 0x00010004L;

  /** root node of the trie */
  private TrieNode root;
  /** size of each node (alphabet size) */
  private int      size;
  /** whether to reverse key direction */
  private boolean  rev;

  /*------------------------------------------------------------------*/
  /** Create a subproblem trie.
   *  @param  vertcnt  the number of perimeter vertices
   *  @param  rev      whether to reverse the key direction   
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Trie (int vertcnt, boolean rev)
  { this.size = vertcnt; this.rev = rev; this.root = new TrieNode(); }

  /*------------------------------------------------------------------*/
  /** Create a subproblem trie.
   *  @param  vertcnt  the number of perimeter vertices
   *  @param  holecnt  the number of hole      vertices
   *  @param  rev      whether to reverse the key direction   
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Trie (int vertcnt, int holecnt, boolean rev)
  { this(vertcnt +holecnt, rev); }

  /*------------------------------------------------------------------*/
  /** Create a subproblem trie.
   *  @param  vertcnt  the number of perimeter vertices
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Trie (int vertcnt)
  { this(vertcnt, false); }

  /*------------------------------------------------------------------*/
  /** Create a subproblem trie.
   *  @param  vertcnt  the number of perimeter vertices
   *  @param  holecnt  the number of hole      vertices
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Trie (int vertcnt, int holecnt)
  { this(vertcnt +holecnt, false); }

  /*------------------------------------------------------------------*/
  /** Add a subproblem to the trie.
   *  @param  key     the key  of the subproblem
   *  @param  weight  the weight of the solution of the subproblem
   *  @param  object  the solution of the subproblem
   *  @return  0 if adding the subproblem was successful,
   *          -1 if the corresponding node os already in use
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int add (Vertex[] key, double weight, Object object)
  { return this.add(key, key.length, weight, object); }

  /*------------------------------------------------------------------*/
  /** Add a subproblem to the trie.
   *  @param  key     the key  of the subproblem
   *  @param  len     the size of the subproblem
   *  @param  weight  the weight of the solution of the subproblem
   *  @param  object  the solution of the subproblem
   *  @return  0 if adding the subproblem was successful,
   *          -1 if the corresponding node os already in use
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int add (Vertex[] key, int len, double weight, Object object)
  {                             /* --- add a subproblem solution */
    int      i, k;              /* vertex identifier, loop variable */
    TrieNode n = this.root;     /* current node in the trie */

    if (this.rev) {             /* if to use the key backwards */
      for (k = len; --k >= 0;){ /* traverse the key */
        if (n.children == null) /* create a child node vector */
          n.children = new TrieNode[this.size+1];
        i = (key[k] != null)    /* get the index of the child */
          ?  key[k].id : this.size;
        if (n.children[i] == null)
          n.children[i] = new TrieNode();
        n = n.children[i];      /* if the child does not exist, */
      } }                       /* create it, then go to it */
    else {                      /* if to use the key forward */
      for (k = 0; k < len; k++){/* traverse the key */
        if (n.children == null) /* create a child node vector */
          n.children = new TrieNode[this.size+1];
        i = (key[k] != null)    /* get the index of the child */
          ?  key[k].id : this.size;
        if (n.children[i] == null)
          n.children[i] = new TrieNode();
        n = n.children[i];      /* if the child does not exist, */
      }                         /* create it, then go to it */
    }
    if (n.weight >= 0)          /* if the node is already in use, */
      return -1;                /* abort the function */
    n.weight = weight;          /* store the triangulation weight */
    n.object = object;          /* and the splitting path/vertex */
    return 0;                   /* and return 'ok' */
  }  /* add() */

  /*------------------------------------------------------------------*/
  /** Get the trie node for a subproblem.
   *  @param  key  the key  of the subproblem
   *  @param  len  the size of the subproblem
   *  @return the trie node for the subproblem
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private TrieNode get (Vertex[] key, int len)
  {                             /* --- get a trie node */
    int      i, k;              /* vertex identifier, loop variable */
    TrieNode n = this.root;     /* current node in the trie */

    if (this.rev) {             /* if to use the key backwards */
      for (k = len; --k >= 0;){ /* traverse the key */
        if ((n == null) || (n.children == null))
          return null;          /* check node and children */
        i = (key[k] != null) ? key[k].id : this.size;
        n = n.children[i];      /* compute the child index */
      } }                       /* and go the child node */
    else {                      /* if to use the key forward */
      for (k = 0; k < len; k++){/* traverse the key */
        if ((n == null) || (n.children == null))
          return null;          /* check node and children */
        i = (key[k] != null) ? key[k].id : this.size;
        n = n.children[i];      /* compute the child index */
      }                         /* and go the child node */
    }
    return n;                   /* return the retrieved node */
  }  /* get() */

  /*------------------------------------------------------------------*/
  /** Get a subproblem solution (triangulation weight).
   *  @param  key  the key of the subproblem
   *  @return the weight of the triangulation
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getWeight (Vertex[] key)
  { return this.getWeight(key, key.length); }

  /*------------------------------------------------------------------*/
  /** Get a subproblem solution (triangulation weight).
   *  @param  key  the key  of the subproblem
   *  @param  len  the size of the subproblem
   *  @return the weight of the triangulation
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public double getWeight (Vertex[] key, int len)
  {                             /* --- get a subproblem solution */
    TrieNode n = this.get(key, len);
    return (n != null) ? n.weight : -1.0;
  }  /* getWeight() */          /* return the weight in the node */

  /*------------------------------------------------------------------*/
  /** Get a subproblem solution.
   *  @param  key  the key of the subproblem
   *  @return the solution of the subproblem
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Object getObject (Vertex[] key)
  { return this.getObject(key, key.length); }

  /*------------------------------------------------------------------*/
  /** Get a subproblem solution.
   *  @param  key  the key  of the subproblem
   *  @param  len  the size of the subproblem
   *  @return the solution of the subproblem
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Object getObject (Vertex[] key, int len)
  {                             /* --- get a subproblem solution */
    TrieNode n = this.get(key, len);
    return (n != null) ? n.object : null;
  }  /* getObject() */          /* return the splitting path/vertex */

  /*------------------------------------------------------------------*/
  /** Count nodes, used nodes, leaves, number of pointers.
   *  @param  n  the root of the current subtree
   *  @param  v  the vector of counters to update
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void count (TrieNode n, int[] v)
  {                             /* --- count nodes */
    int i;                      /* loop variable */

    v[0]++;                     /* count nodes, used nodes, leaves */
    if (n.weight  >= 0)     v[1]++;
    if (n.children == null) v[3]++;
    else {                      /* traverse the child nodes */
      v[4] += this.size +1;     /* count the number of pointers */
      for (i = 0; i <= this.size; i++)
        if (n.children[i] != null)
          this.count(n.children[i], v);
    }                           /* count nodes recursively */
  }  /* count() */

  /*------------------------------------------------------------------*/
  /** Get statistics about the trie.
   *  @return a list of five integers with different counts
   *  @since  2005.02.18 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public int[] getStats ()
  {                             /* --- get statistics */
    int[] v = new int[5];       /* create a counter vector */
    this.count(this.root, v);   /* count the nodes recursively */
    v[2] = v[0] -v[3];          /* compute number of inner nodes */
    return v;                   /* return the node counters */
  }  /* showStats() */

}  /* class Trie */
